#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine operator mapping file."""

from .op import Operator, operator_registry
from .tensor import Tensor
import numpy as np


@operator_registry(operator_type='Gather')
class Gather(Operator):
    """Parse the Gather operator to the neural engine."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def set_attr(self, framework, node):
        """Extract the node attr from frameworks."""
        self._op_type = 'Gather'
        if framework == 'tensorflow':
            self._attr['batch_dims'] = node.attr['batch_dims'].i
            try:
                axis = int(self._input_tensors[2].data)
            except BaseException:
                axis = 0
            self._attr['axis'] = axis
            
        if framework == 'onnxruntime':
            # idx_axis
            self._attr['axis'] = 0
            if len(node.attribute) != 0:
                # src_axis
                self._attr['batch_dims'] = node.attribute[0].i
            else:
                self._attr['batch_dims'] = 0
            if isinstance(self._input_tensors[1].data, np.ndarray) and \
               len(self._input_tensors[1].data.shape) == 0:
                self._attr['keep_dims'] = False


                self._attr['axis'] = 0
            
        if framework == 'torch':
            pass
# tf.gather(params, indices, validate_indices=None, axis=None, batch_dims=0, name=None)
# argument validate_indices is deprecated
# indices must be an integer tensor of any dimension (usually 0-D or 1-D).
# Produces an output tensor with shape
# params.shape[:axis] + indices.shape[batch_dims:] + params.shape[axis + 1:]
# see: https://www.tensorflow.org/api_docs/python/tf/gather
@operator_registry(operator_type='GatherV2')
class GatherV2(Operator):
    """Parse the GatherV2 operator to the neural engine."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def set_attr(self, framework, node):
        """Extract the node attr from frameworks."""
        self._op_type = 'Gather'
        if framework == 'tensorflow':
            self._attr['batch_dims'] = node.attr['batch_dims'].i
            try:
                axis = int(self._input_tensors[2].data)
            except BaseException:
                axis = 0
            _ = self._input_tensors.pop()
            self._attr['axis'] = axis



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The OperatorAdaptor pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
from ..ops.tensor import Tensor
import numpy as np
import copy


@pattern_registry(pattern_type='OperatorAdaptor')
class OperatorAdaptor(Pattern):

    """The OperatorAdaptor pattern.

    Modify some operators after all fusion patterns. For example, sweep input tensors to make
    sure that Neural Engine backend op can receive its required inputs in different framework.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class.

        Args:
            model (class): the Graph class
            sweep_nodes_info (dict): Supply the handling information for model to sweep inputs.
                                     The key (str) is the node name and the value (str) is
                                     the target position.
        """
        def _sweep_inputs(model, sweep_nodes_info):
            for name, target_pos in sweep_nodes_info.items():
                node_id = model.get_node_id(name)
                target_inputs = [model.nodes[node_id].input_tensors[j] for j in target_pos]
                model.nodes[node_id].input_tensors = target_inputs

        def _reshape_non_2d_src_before_inner_product(model, node_names):
            for name in node_names:
                node_id = model.get_node_id(name)
                ip_node = model.nodes[node_id]
                hidden_size = int(ip_node.input_tensors[1].data.shape[0])
                pre_node = model.get_node_by_name(ip_node.input_tensors[0].source_op[0])
                reshape_node_name = "reshape_2d_" + name
                pre_node.output_tensors[0].dest_op.append(reshape_node_name)
                input_tensors = [copy.deepcopy(pre_node.output_tensors[0])]
                output_tensors = [Tensor(name=reshape_node_name+":0",
                    source_op=[reshape_node_name], dest_op=[name])]
                ip_node.input_tensors[0] = output_tensors[0]
                reshape_node = util.construct_node(reshape_node_name, 'Reshape',
                    input_tensors=input_tensors, output_tensors=output_tensors, attr=OrderedDict(
                    {'dst_shape': '-1,' + str(hidden_size)}))
                model.insert_nodes(node_id, [reshape_node])

        # k: op_type, v :target_pos
        sweep_op = {'Gather': [1,0]}
        sweep_nodes_info = {}
        ip_need_reshape_2d_names = []
        for n in model.nodes:
            if n.op_type in sweep_op.keys():
                # skip the Gather op which has reqiured inputs
                if n.op_type == 'Gather' and n.filling_method == 'extract_from_onnxruntime':
                    sweep_nodes_info[n.name] = sweep_op[n.op_type]
            elif n.op_type in ['MatMul', 'MatMulWithBias', 'MatMulWithBiasAdd']:
                if isinstance(n.input_tensors[1].data, np.ndarray):
                    pre_node = None
                    try:
                        pre_node = model.get_node_by_name(n.input_tensors[0].source_op[0])
                    except:
                        pre_node = None
                    if pre_node and pre_node.op_type == 'Transpose' and \
                       len(pre_node.attr['dst_perm']) > 3:
                        ip_need_reshape_2d_names.append(n.name)
            else:
                continue
        if len(sweep_nodes_info) > 0:
            _sweep_inputs(model, sweep_nodes_info)
        if len(ip_need_reshape_2d_names) > 0:
            _reshape_non_2d_src_before_inner_product(model, ip_need_reshape_2d_names)

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The MatMulWithTranspose pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
from ..onnx_utils import bias_to_int32
import copy


@pattern_registry(pattern_type='MatMulWithTranspose')
class MatMulWithTranspose(Pattern):
    """The MatMulWithTranspose pattern.

    Fuse the original sub-graph into the custom acceleration 'MatMulWithTranspose' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'MatMulWithTranspose': [
                {
                    'patterns': {
                        'in': [[(0, 'Reorder'), (1, 'Reorder'), (3, 'Matmul')],
                               [(), (2, 'Reorder'), (3, 'Matmul')]
                               ],
                        'out': [[(0, 'MatmulwithTranspose')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 3
                    },
                    'input_tensors': {
                        0: [[{
                            2: [0]
                        }, {
                            0: [0]
                        }], [[0, 1], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            3: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 1, 2]
                },
                
                {
                    'patterns': {
                        'in': [[(0, 'Reorder'), (1, 'Matmul'), (2, 'Reorder'), (3, 'Shape'), (5, 'View')],
                               [(2, 'Reorder'), (4, 'Shape'), (5, 'View')]
                               ],
                        'out': [[(0, 'Matmul')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 1
                    },
                    'input_tensors': {
                        0: [[{
                            1: [0]
                        }, {
                            0: [0]
                        }], [[0, 1], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            5: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 2, 5]
                },
                
                {
                    'patterns': {
                        'in': [[(0, 'Reorder'), (4, 'Matmul')],
                               [(), (1, 'Reorder'), (2, 'Concat'), (3, 'Reorder'), (4, 'Matmul')]
                               ],
                        'out': [[(0, 'Concat'), (1, 'MatmulwithTranspose')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 2,
                        1: 4,
                    },
                    'input_tensors': {
                        0: [[{
                            2: [0]
                        }, {
                            1: [0]
                        }], [[0, 1], 2]],
                        1: [[{
                            0: [0]
                        }], [[0], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            2: [0]
                        }], [[0], 1]],
                        1: [[{
                            4: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 3, 1]
                },

                {
                    'patterns': {
                        'in': [[(0, 'Reorder'), (1, 'Concat'), (2, 'Matmul'), (3, 'Reorder'),
                                (4, 'Shape'), (6, 'View')],
                               [(3, 'Reorder'), (5, 'Shape'), (6, 'View')]
                               ],
                        'out': [[(0, 'Concat'), (1, 'MatmulwithTranspose')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 1,
                        1: 2,
                    },
                    'input_tensors': {
                        0: [[{
                            1: [0]
                        }, {
                            0: [0]
                        }], [[0, 1], 2]],
                        1: [[{
                            2: [0]
                        }], [[0], 2]]
                    },
                    'output_tensors': {
                         0: [[{
                            1: [0]
                        }], [[0], 1]],
                        1: [[{
                            6: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 2, 3, 6]
                },
            ]
        }
        if model.framework_modeling_config['framework'] != 'torch':
            return model
        def _set_attr(new_node_names, ret_old_nodes, model):
            for i in range(len(new_node_names)):
                transpose_a = ret_old_nodes[i][0].attr['dst_perm']
                transpose_b = ret_old_nodes[i][2].attr['dst_perm']
                a_trans = ret_old_nodes[i][1].attr['transpose_dims']
                a_trans = util.str2list(a_trans)
                mat_node_idx = model.get_node_id(new_node_names[i][0])
                attr = OrderedDict()
                if transpose_a:
                    attr['src1_perm'] = transpose_a
                    if a_trans:
                        transpose_a = util.str2list(transpose_a)
                        tmp = transpose_a[a_trans[0]]
                        transpose_a[a_trans[0]] = transpose_a[a_trans[1]]
                        transpose_a[a_trans[1]] = tmp
                        attr['src1_perm'] = util.list2str(transpose_a)
                if transpose_b:
                    attr['src0_perm'] = transpose_b
                model.nodes[mat_node_idx].attr = attr
                
                concat_node =  model.get_node_by_name(model.nodes[mat_node_idx].input_tensors[0].source_op[0])
                concat1_node = model.get_node_by_name(model.nodes[mat_node_idx].input_tensors[1].source_op[0])
                if concat_node.op_type == "Concat":
                    concat_node.attr = OrderedDict({'axis': '3'})
                if concat1_node.op_type == "Concat":
                    concat1_node.attr = OrderedDict({'axis': '3'})    

        pattern_dict = pattern_mapping_config['MatMulWithTranspose'][0]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("MatMulWithTranspose", 
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            _set_attr(new_node_names, ret_old_nodes, model)


        pattern_dict = pattern_mapping_config['MatMulWithTranspose'][1]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("MatMulWithTranspose", 
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            for i in range(len(new_node_names)):
                transpose_a = ret_old_nodes[i][0].attr['dst_perm']
                transpose_b = ret_old_nodes[i][1].attr['dst_perm']
                mat_node_idx = model.get_node_id(new_node_names[i][0])
                attr = OrderedDict()
                if transpose_a:
                    attr['src1_perm'] = transpose_a
                if transpose_b:
                    attr['dst_perm'] = transpose_b
                reshape_attr = util.str2list(ret_old_nodes[i][2].attr['shape'])
                if reshape_attr:
                    attr['reshape'] = '-1, ' + str(reshape_attr[-1])
                model.nodes[mat_node_idx].attr = attr
                
                
                
        def _set_attr1(new_node_names, ret_old_nodes, model):
            for i in range(len(new_node_names)):
                transpose_a = ret_old_nodes[i][0].attr['dst_perm']
                transpose_dims = util.str2list(ret_old_nodes[i][1].attr.get('transpose_dims',
                                                                            '-1,-2'))
                assert len(transpose_dims) == 2
                b_dim_ori = util.str2list(ret_old_nodes[i][2].attr['dst_perm'])
                b_dim_ori[transpose_dims[0]], b_dim_ori[transpose_dims[1]] = \
                    b_dim_ori[transpose_dims[1]], b_dim_ori[transpose_dims[0]]
                transpose_b = util.list2str(b_dim_ori) #"0,1,3,2"
                # a_trans = ret_old_nodes[i][1].attr['transpose_dims']
                # a_trans = util.str2list(a_trans)
                mat_node_idx = model.get_node_id(new_node_names[i][1])
                attr = OrderedDict()
                if transpose_a:
                    attr['src0_perm'] = transpose_a
                    # if a_trans:
                    #     transpose_b = util.str2list(transpose_a)
                    #     tmp = transpose_a[a_trans[0]]
                    #     transpose_a[a_trans[0]] = transpose_a[a_trans[1]]
                    #     transpose_a[a_trans[1]] = tmp
                    #     attr['src1_perm'] = util.list2str(transpose_a)
                if transpose_b:
                    attr['src1_perm'] = transpose_b
                model.nodes[mat_node_idx].attr = attr
                
                concat_node =  model.get_node_by_name(model.nodes[mat_node_idx].input_tensors[0].source_op[0])
                concat1_node = model.get_node_by_name(model.nodes[mat_node_idx].input_tensors[1].source_op[0])
                if concat_node.op_type == "Concat":
                    concat_node.attr = OrderedDict({'axis': '3'})
                if concat1_node.op_type == "Concat":
                    concat1_node.attr = OrderedDict({'axis': '1'})
                    concat2 = model.get_node_by_name(concat1_node.input_tensors[1].source_op[0])
                    # concat2 = model.get_node_by_name(reorder_node1.input_tensors[0].source_op[0])
                    concat2.attr = OrderedDict({'axis': '3'})
                    
        pattern_dict = pattern_mapping_config['MatMulWithTranspose'][2]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("MatMulWithTranspose", 
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            _set_attr1(new_node_names, ret_old_nodes, model)


        pattern_dict = pattern_mapping_config['MatMulWithTranspose'][3]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("MatMulWithTranspose", 
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            for i in range(len(new_node_names)):

                mat_node_idx = model.get_node_id(new_node_names[i][1])
                attr = OrderedDict()
                transpose_b = ret_old_nodes[i][0].attr['dst_perm']
                transpose_dst = ret_old_nodes[i][2].attr['dst_perm']
                if transpose_b:
                    attr['src1_perm'] = transpose_b
                if transpose_dst:
                    attr['dst_perm'] = transpose_dst
                reshape_attr = util.str2list(ret_old_nodes[i][3].attr['shape'])
                if reshape_attr:
                    attr['reshape'] = '-1, ' + str(reshape_attr[-1])
                model.nodes[mat_node_idx].attr = attr
                concat_node = model.get_node_by_name(model.nodes[mat_node_idx].input_tensors[1].source_op[0])
                concat_node.attr = OrderedDict({'axis': '1'})
                
            return model
        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The TextEncoder_SoftmaxReshape Pattern."""

from .pattern import Pattern, pattern_registry
from collections import OrderedDict
import copy
from .. import graph_utils as util
from .. import logger


@pattern_registry(pattern_type='TextEncoder_SoftmaxReshape')
class TextEncoder_SoftmaxReshape(Pattern):
    """The TextEncoder_SoftmaxReshape pattern.

    Fuse the original sub-graph into the custom acceleration 'TextEncoder_SoftmaxReshape' graph.
    The search strategy is based on the following pattern mapping configs for the stable diffusionV1-5.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            # for text encoder self_attn/out_proj/Add Reshape, Reshape 7
            'TextEncoder_SoftmaxReshape': [
                {
                    'patterns': {
                        'in': [[(0, 'Reshape'), (1, 'Shape'), (2, 'Gather'), (3, 'Unsqueeze'), (6, 'Concat'),
                                (7, 'Reshape'), (8, 'Softmax')], [(), (4, 'Unsqueeze'), (6, 'Concat')],
                               [(), (5, 'Unsqueeze'), (6, 'Concat')]],
                        'out': [[(0, 'Reshape'), (1, 'Reshape'), (2, 'Softmax')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0,
                        1: 7,
                        2: 8
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            0: [1]
                        }], [[0, 1], 2]],
                        1: [[{
                            7: [0]
                        }], [[0], 2]],
                        2: [[], [[], 1]],
                    },
                    'output_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]],
                        1: [[{
                            7: [0]
                        }], [[0], 1]],
                        2: [[{
                            8: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 6, 8]
                },
            ]
        }

        def _set_attr(node_names, model):
            attr = OrderedDict()
            mul_size = 12
            # mul_size + max_seq + seq 12,77,77
            attr['dst_shape'] = str(mul_size) + ',-1,-1'
            attr['dims'] = '1,1'
            reshape_node_idx = model.get_node_id(node_names[1])
            model.nodes[reshape_node_idx].attr = attr

        for i in range(len(pattern_mapping_config['TextEncoder_SoftmaxReshape'])):
            pattern_dict = pattern_mapping_config['TextEncoder_SoftmaxReshape'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("TextEncoder_SoftmaxReshape",
                                                                        pattern_dict, model)

            logger.info('TextEncoder_SoftmaxReshape mathched...')
            logger.debug('TextEncoder_SoftmaxReshape = {}'.format(new_node_names))
            if len(new_node_names) != 0:
                for j in range(len(new_node_names)):
                    first_ret_old_node = ret_old_nodes[j][0]
                    assert first_ret_old_node.op_type == 'Reshape'
                    reshape_new_node_idx = model.get_node_id(new_node_names[j][0])
                    model.nodes[reshape_new_node_idx].attr = first_ret_old_node.attr
                    _set_attr(new_node_names[j], model)

                    transpose_node_name = model.nodes[reshape_new_node_idx].output_tensors[0].dest_op[1]
                    transpose_node_idx = model.get_node_id(transpose_node_name)

                    # delete the node before insert
                    copy_node = copy.deepcopy(model.nodes[reshape_new_node_idx])
                    model.remove_nodes([new_node_names[j][0]])
                    model.insert_nodes(transpose_node_idx, [copy_node])

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The AttentionReshape Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
import copy
from .. import graph_utils as util


@pattern_registry(pattern_type='AttentionReshape')
class AttentionReshape(Pattern):
    """The AttentionReshape pattern.

    Fuse the original sub-graph into the custom acceleration 'AttentionReshape' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'AttentionReshape': [
                {
                    'patterns': {
                        'in': [[(0, 'Mul'), (1, 'Pack'), (2, 'Reshape')]],
                        'out': [[(0, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 2
                    },
                    'input_tensors': {
                        0: [[{
                            2: [0]
                        }], [[0], 1]]
                    },
                    'output_tensors': {
                        0: [[{
                            2: [0]
                        }], [[0], 1]]
                    },
                    'returns': [1]
                },

                # bert_base_sparse
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'), (6, 'Concat'),
                                (7, 'Reshape'), (8, 'MatMulWithBias')],
                                [(), (3, 'Shape'), (4, 'Gather'), (5, 'Unsqueeze'),
                                (6, 'Concat')]],
                        'out': [[(0, 'Reshape'), (1, 'MatMulWithBias')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 7,
                        1: 8
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]],
                        1: [[
                            {8: [1]}, {8: [2]}
                        ], [[1, 2], 3]],
                    },
                    'output_tensors': {
                        0: [[{
                            7: [0]
                        }], [[0], 1]],
                        1: [[{
                            8: [0]
                        }], [[0], 1]]
                    },
                    'returns': [6, 8]
                },

                # Lat_int8
                {
                    'patterns': {
                        'in': [ [(0, 'Shape'), (1, 'Gather'), (2, 'Gather'), (3, 'Unsqueeze'), 
                                (4, 'Concat'), (6, 'Reshape'),(7, 'MatMulWithBias')],
                                [(),(5, 'Transpose'), (6, 'Reshape')]],
                        'out': [[(0, 'Transpose'), (1, 'Reshape'), (2, 'MatMulWithBias')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 5,
                        1: 6,
                        2: 7,
                    },
                    'input_tensors': {
                        0: [[{
                            5: [0]
                        }], [[0], 1]],
                        1: [[], [[],1]],
                        2: [[
                        {7: [1]},{7: [2]}
                        ],[[1, 2],3]],
                    },
                    'output_tensors': {
                        0: [[], [[],1]],

                        1: [[{
                        6:[0]
                        }],[[0],1]],
                        2: [[{
                            7: [0]
                        }], [[0], 1]]
                    },
                    'returns': [5,4,7]
                },

                # Shira new model Reshape_128, 373, 618
                {
                    'patterns': {
                        'in': [[(0, ['MatMul', 'BatchMatMul']), (1, 'Transpose'), (2, 'Shape'),
                                (3, 'Gather'), (4, 'Unsqueeze'), (5, 'Concat'), (6, 'Reshape'),
                                (7, 'MatMulWithBias')]
                                ],
                        'out': [[(0, 'BatchMatMul'), (1, 'Transpose'), (2, 'Reshape'),
                                 (3, 'MatMulWithBias')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0,
                        1: 1,
                        2: 6,
                        3: 7,
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            0: [1]
                        }], [[0, 1], 2]],
                        1: [[], [[], 1]],
                        2: [[], [[], 1]],
                        3: [[{
                            7: [1]
                        },{
                            7: [2]
                        }],[[1, 2], 3]],
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[{
                            6: [0]
                        }],[[0], 1]],
                        3: [[{
                            7: [0]
                        }], [[0], 1]]
                    },
                    'returns': [1, 5, 7, 0]
                },

                # bert_mini_int8
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Gather'), (3, 'Unsqueeze'), (6, 'Concat'),
                                (7, 'Reshape'), (8, 'MatMulWithBias')],
                                [(1, 'Gather'), (4, 'Gather'), (5, 'Unsqueeze'),
                                (6, 'Concat')]],
                        'out': [[(0, 'Reshape'), (1, 'MatMulWithBias')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 7,
                        1: 8
                    },
                    'input_tensors': {
                        0: [[{
                            7: [0]
                        }], [[0], 1]],
                        1: [[
                            {8: [1]}, {8: [2]}
                        ], [[1, 2], 3]],
                    },
                    'output_tensors': {
                        0: [[{
                            7: [0]
                        }], [[0], 1]],
                        1: [[{
                            8: [0]
                        }], [[0], 1]]
                    },
                    'returns': [6, 8]
                },

                # distil_bert_base
                {
                    'patterns': {
                        'in': [[(0, 'Unsqueeze'), (1, 'Concat'), (2, 'Reshape'),
                                (3, 'MatMulWithBias')]],
                        'out': [[(0, 'Reshape'), (1, 'MatMulWithBias')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 2,
                        1: 3
                    },
                    'input_tensors': {
                        0: [[{
                            2: [0]
                        }], [[0], 1]],
                        1: [[
                            {3: [1]}, {3: [2]}
                        ], [[1, 2], 3]],
                    },
                    'output_tensors': {
                        0: [[{
                            2: [0]
                        }], [[0], 1]],
                        1: [[{
                            3: [0]
                        }], [[0], 1]]
                    },
                    'returns': [1, 3]
                },

                # geminet
                {
                    'patterns': {
                        'in': [[(0, 'Reshape'), (1, 'MatMulWithBias')]],
                        'out': [[(0, 'Reshape'), (1, 'MatMulWithBias')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0,
                        1: 1
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]],
                        1: [[
                            {1: [1]}, {1: [2]}
                        ], [[1, 2], 3]],
                    },
                    'output_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]],
                        1: [[{
                            1: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 1]
                },
            ]
        }

        def _set_attr(hidden_size, node_names, model, reshape_idx=0):
            attr = OrderedDict()
            attr['dst_shape'] = '-1,' + str(hidden_size)
            reshape_node_idx = model.get_node_id(node_names[reshape_idx])
            model.nodes[reshape_node_idx].attr = attr

        for i in range(len(pattern_mapping_config['AttentionReshape'])-1):
            pattern_dict = pattern_mapping_config['AttentionReshape'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("AttentionReshape",
                                                                        pattern_dict, model)

            if len(new_node_names) != 0:
                for j in range(len(new_node_names)):
                    if len(ret_old_nodes[j]) == 3 or len(ret_old_nodes[j]) == 4:
                        pack_node = ret_old_nodes[j][1]
                        hidden_size = int(pack_node.input_tensors[-1].data)
                        reshape_idx = 1 if len(ret_old_nodes[j]) == 3 else 2
                        _set_attr(hidden_size, new_node_names[j], model,reshape_idx=reshape_idx)

                        assert ret_old_nodes[j][0].op_type == 'Transpose'
                        idx_get = 0 if len(ret_old_nodes[j]) == 3 else 1
                        trans_node_idx = model.get_node_id(new_node_names[j][idx_get])
                        model.nodes[trans_node_idx].attr = ret_old_nodes[j][0].attr

                        assert ret_old_nodes[j][2].op_type == 'MatMulWithBias'
                        idx_get = 2 if len(ret_old_nodes[j]) == 3 else 3
                        mat_node_idx = model.get_node_id(new_node_names[j][idx_get])
                        model.nodes[mat_node_idx].attr = ret_old_nodes[j][2].attr

                        if ret_old_nodes[j][-1].op_type == 'BatchMatMul':
                            bmat_node_idx = model.get_node_id(new_node_names[j][0])
                            model.nodes[bmat_node_idx].attr = ret_old_nodes[j][-1].attr

                    elif len(ret_old_nodes[j]) == 2:
                        pack_node = ret_old_nodes[j][0]
                        hidden_size = int(pack_node.input_tensors[-1].data)
                        _set_attr(hidden_size, new_node_names[j], model)
                        assert ret_old_nodes[j][1].op_type == 'MatMulWithBias'
                        mat_node_idx = model.get_node_id(new_node_names[j][1])
                        model.nodes[mat_node_idx].attr = ret_old_nodes[j][1].attr
                return model
        # special reshape node, like has '0,0,768' or '-1,369,384' dst_shape attr
        pattern_dict = pattern_mapping_config['AttentionReshape'][-1]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("AttentionReshape",
                                                                    pattern_dict, model)

        if len(new_node_names) != 0:
            for j in range(len(new_node_names)):
                reshape_node = ret_old_nodes[j][0]
                dst_shape = reshape_node.attr.get('dst_shape', None)
                if dst_shape != None and dst_shape.split(',')[0] == '0' or '-1':
                    hidden_size = int(dst_shape.split(',')[-1])
                    _set_attr(hidden_size, new_node_names[j], model)
                mat_node_idx = model.get_node_id(new_node_names[j][1])
                model.nodes[mat_node_idx].attr = ret_old_nodes[j][1].attr

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The AttentionMaskLengthAdaptiveExpandIndices Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
import numpy as np


@pattern_registry(pattern_type='AttentionMaskLengthAdaptiveExpandIndices')
class AttentionMaskLengthAdaptiveExpandIndices(Pattern):
    """The AttentionMaskLengthAdaptiveExpandIndices pattern.

    Fuse the original sub-graph into the custom acceleration 'AttentionMaskLengthAdaptiveExpandIndices' graph.
    The fusion strategy is based on 'AddClsToken' pattern map configurations and different kinds of models.
    """

    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'AttentionMaskLengthAdaptiveExpandIndices': [
                # minilmv2-lat-roberta
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'), (9, 'Concat'),
                                (10, 'Reshape'), (11, 'Shape'), (12, 'ConstantOfShape'),
                                (13, 'Mul'), (14, 'Equal'), (15, 'Where'), (18, 'Expand')],
                               [(), (3, 'Shape'), (4, 'Gather'), (5, 'Unsqueeze'), (9, 'Concat')],
                               [(), (6, 'Shape'), (7, 'Gather'), (8, 'Unsqueeze'), (9, 'Concat')],
                               [(), (16, 'Unsqueeze'), (17, 'Unsqueeze'), (18, 'Expand')]],
                        'out': [[(0, 'ExpandIndices')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 18,
                    },
                    'input_tensors': {
                        0: [[{
                            16: [0]
                        }, {
                            0: [0]
                        }], [[0, 1], 2]],
                    },
                    'output_tensors': {
                        0: [[{
                            18: [0]
                        }], [[0], 1]],
                    },
                    'returns': [16, 17]
                },
                # Expand 229
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'), (6, 'Concat'),
                                (7, 'Reshape'), (8, 'Shape'), (9, 'ConstantOfShape'),
                                (10, 'Mul'), (11, 'Equal'), (12, 'Where'), (15, 'Expand')],
                               [(), (3, 'Shape'), (4, 'Gather'), (5, 'Unsqueeze'), (6, 'Concat')],
                               [(), (13, 'Unsqueeze'), (14, 'Unsqueeze'), (15, 'Expand')]],
                        'out': [[(0, 'ExpandIndices')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 15,
                    },
                    'input_tensors': {
                        0: [[{
                            13: [0]
                        }, {
                            0: [0]
                        }], [[0, 1], 2]],
                    },
                    'output_tensors': {
                        0: [[{
                            15: [0]
                        }], [[0], 1]],
                    },
                    'returns': [13, 14]
                },
                # minilmv2-lat-roberta int8
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'), (5, 'Concat'),
                                (6, 'Reshape'), (7, 'Equal'), (8, 'Where'),  (11, 'Expand')],
                               [(0, 'Shape'), (3, 'Gather'), (4, 'Unsqueeze'), (5, 'Concat')],
                               [(), (9, 'Unsqueeze'), (10, 'Unsqueeze'), (11, 'Expand')]],
                        'out': [[(0, 'ExpandIndices')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 11,
                    },
                    'input_tensors': {
                        0: [[{
                            9: [0]
                        }, {
                            0: [0]
                        }], [[0, 1], 2]],
                    },
                    'output_tensors': {
                        0: [[{
                            11: [0]
                        }], [[0], 1]],
                    },
                    'returns': [9, 10]
                },
            ]
        }

        # minilmv2-lat-roberta
        for idx, pattern_dict in enumerate(pattern_mapping_config['AttentionMaskLengthAdaptiveExpandIndices']):
            model, new_node_names, ret_old_nodes = \
             util.pattern_mapping('AttentionMaskLengthAdaptiveExpandIndices', pattern_dict, model)
            if len(new_node_names) != 0:
                for i in range(len(new_node_names)):
                    attr = OrderedDict()
                    input_indices = []
                    for unsqueeze_node in ret_old_nodes[i]:
                        input_indices.append(int(unsqueeze_node.attr['axes']))
                        attr['position'] = util.list2str(input_indices)
                        keep_indices_node_idx = model.get_node_id(new_node_names[i][0])
                        model.nodes[keep_indices_node_idx].attr = attr

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Data augmentation API.

Please refer to https://github.com/intel/intel-extension-for-transformers/blob/main/docs/data_augmentation.md.
"""

import csv
import json
import math
import numpy as np
import os
from datasets import load_dataset
from intel_extension_for_transformers.optimization.utils.utility import LazyImport
from operator import methodcaller
from tqdm import tqdm
from .utils import AugmenterType, get_augmenter_from_type

torch = LazyImport("torch")

DEFAULT_OUTPUT_FILE = "augmented_dataset"


class DataAugmentation:
    """DataAugmentation provides many ways to enhance existing datasets.

    Args:
            augmenter_type (string): augmenter type like: "TextGenerationAug", "KeyboardAug", ...

    Example::

        aug = DataAugmentation(augmenter_type="TextGenerationAug")
        aug.input_dataset = self.origin_data
        aug.output_path = os.path.join(self.result_path, "test1.cvs")
        aug.augmenter_arguments = {'model_name_or_path': 'gpt2-medium'}
        aug.data_augment()
    """
    def __init__(self, augmenter_type: str):
        """Init an instance of DataAugmentation."""
        self._augmenter_type = AugmenterType[augmenter_type.upper()].value
        self._output_path = "save_path/augmented_dataset.csv"
        self._input_dataset = None
        self._data_config_or_task_name = None
        self._augmenter_arguments = None
        self._column_names = "sentence"
        self._split = "validation"
        self._num_samples = 1
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    @property
    def augmenter_type(self):
        """Get augmenter type."""
        return self._augmenter_type

    @augmenter_type.setter
    def augmenter_type(self, augmenter_type):
        # pylint: disable=C0301
        """Set augmenter type.

        Please refer to https://github.com/intel/intel-extension-for-transformers/blob/main/docs/data_augmentation.md#supported-augmenter
        """
        self._augmenter_type = AugmenterType[augmenter_type.upper()].value

    @property
    def output_path(self):
        """Get output path for new datasets."""
        return self._output_path

    @output_path.setter
    def output_path(self, output_path):
        """Set output path for new datasets."""
        self._output_path = output_path

    @property
    def input_dataset(self):
        """Get path for original datasets."""
        return self._input_dataset

    @input_dataset.setter
    def input_dataset(self, input_dataset):
        # pylint: disable=C0301
        """Set original datasets.

        Input_dataset can be a file path, the file should be contained the samples as format: 'label' + '\t' + 'sentence' + EOS + '\n',please refer to
        https://github.com/intel/intel-extension-for-transformers/blob/1d20bcafc3a1f5953400c0c42dafbaa3f700ac1d/tests/test_data_augmentation.py#L9.
        Input_dataset also can be a huggingface datasets name, like: "glue"...,
        if the datasets has task name or data config, you should set data_config_or_task_name:
            example::

                aug.input_dataset = "glue"
                aug.data_config_or_task_name = "sst2"
        """
        self._input_dataset = input_dataset

    @property
    def data_config_or_task_name(self):
        """Get task name for glue datasets or datasets configuration."""
        return self._data_config_or_task_name

    @data_config_or_task_name.setter
    def data_config_or_task_name(self, data_config_or_task_name):
        """Set task name for glue datasets or datasets configuration."""
        self._data_config_or_task_name = data_config_or_task_name

    @property
    def column_names(self):
        """Get column name of datasets."""
        return self._column_names

    @column_names.setter
    def column_names(self, column_names: str):
        """Set column name of datasets."""
        self._column_names = column_names

    @property
    def num_samples(self):
        """Get number of samples generated by augmenter for every original sample."""
        return self._num_samples

    @num_samples.setter
    def num_samples(self, num_samples: int):
        """Set number of samples generated by augmenter for every original sample."""
        self._num_samples = num_samples

    @property
    def augmenter_arguments(self):
        """Get the parameters of augmenter."""
        return self._augmenter_arguments

    @augmenter_arguments.setter
    def augmenter_arguments(self, augmenter_arguments):
        # pylint: disable=C0301
        """Set parameters dict for augmenters. Different augmenter has different parameters.

        For TextGenerationAug type, it is huggingface model name or a path, example: {'model_name_or_path': 'gpt2-medium'}.
        For KeyboardAug type, it is parameters which defined in ["KeyboardAug"](https://github.com/makcedward/nlpaug/blob/40794970124c26ce2e587e567738247bf20ebcad/nlpaug/augmenter/char/keyboard.py#L46).
            example::

                {"aug_char_min": 1, "aug_char_max": 10, "aug_char_p": 0.3, "aug_word_p": 0.3, "aug_word_min": 1, "aug_word_max": 10,
                 "stopwords": None, "tokenizer": None, "reverse_tokenizer": None, "include_special_char": True, "include_numeric": True,
                 "include_upper_case": True, "lang": "en", "verbose": 0, "stopwords_regex": None, "model_path": None, "min_char": 4}
        Please refer [data_augmentation document](https://github.com/intel/intel-extension-for-transformers/blob/main/docs/data_augmentation.md#supported-augmenter)
        """
        self._augmenter_arguments = augmenter_arguments

    @property
    def custom_augmenter(self):
        """Get augmenter which be setted by customer."""
        return self._custom_augmenter

    @custom_augmenter.setter
    def custom_augmenter(self, custom_augmenter):
        """Set customer data augmenter."""
        self._custom_augmenter = custom_augmenter

    def data_augment(self):
        """Execute the process of data augmentation."""
        assert self._input_dataset is not None, \
            "Please pass the dataset name or " \
            "A csv or a json file to DataAugmentation.input_dataset."

        assert self._column_names is not None, \
            "Please pass column names " \
            "which you want to augmentation to DataAugmentation.column_names"

        extension = None
        if os.path.isfile(self._input_dataset):
            extension = self._input_dataset.split(".")[-1]
            assert extension in ["csv", "json"], "`input dataset` should be a csv or a json file."
            if self._input_dataset.endswith(".csv"):
                # Loading a dataset from local csv files
                raw_datasets = load_dataset("csv", data_files=self._input_dataset, delimiter="\t", split="train")
            else:   # pragma: no cover
                # Loading a dataset from local json files
                raw_datasets = load_dataset("json", data_files=self._input_dataset)
        else:   # pragma: no cover
            if self._input_dataset == "glue":
                assert self._data_config_or_task_name is not None, \
                    "Please pass the task name to DataAugmentation.data_config_or_task_name."
            # Downloading and loading a dataset from the hub.
            raw_datasets = load_dataset(
                self._input_dataset, self._data_config_or_task_name, split=self._split
            )
        if extension is None:   # pragma: no cover
            extension = "csv"

        if os.path.isfile(self._output_path):
            self._output_path = self._output_path
        elif os.path.isdir(self._output_path):
            self._output_path = os.path.join(
                self._output_path, DEFAULT_OUTPUT_FILE + "." + extension
            )
        else:
            path, name = os.path.split(self._output_path)
            os.makedirs(path, exist_ok=True)
            self._output_path = \
                os.path.join(
                    path, name if name is not None else DEFAULT_OUTPUT_FILE + "." + extension
                )

        if self._augmenter_type == AugmenterType.TEXTGENERATIONAUG.value:
            self.text_generation_augmentation(extension, raw_datasets)
        else:
            self.mit_data_augmentation(extension, raw_datasets)

    def text_generation_augmentation(self, extension, raw_datasets):
        """Execute the process of text generation augmentation.

        Args:
            extension: No used 
            raw_datasets: The original datasets, the datasets can be from huggingface datasets(like: glue/sst2) or
            the customer datasets, each sample should be:
                'label' + '\t' + 'sentence' + EOS + '\n'

        augmenter_arguments for text generation augmentation::

            {'model_name_or_path': 'gpt2'," \
             'k': 0, // top_k, default: 0
             'p': 0.9, // top_p, default: 0.9
             'temperature': 1.0, // temperature of 1.0 has no effect,
                                 // lower tend toward greedy sampling, default: 1.0
             'repetition_penalty': 1.0, // primarily useful for CTRL model, default: 1.0
             'num_return_sentences': -1 // number of sentences to generate. default is -1 means the entire dataset
             'num_samples': 1, // number of samples generated by each conditional text
                               // generation run. default: 1
             'stop_token': EOS // end of sentence token, at which text generation is stopped.
                              // default: EOS
            }
        """
        assert self._augmenter_arguments is not None, \
            "Please pass a pretrained model name or path to " \
            "DataAugmentation.augmenter_arguments like: " \
            "{'model_name_or_path': 'gpt2'," \
            "......}"
        from transformers import (
            AutoConfig,
            AutoModelForCausalLM,
            AutoTokenizer,
            CTRLLMHeadModel,
            CTRLTokenizer,
            GPT2LMHeadModel,
            GPT2Tokenizer,
            OpenAIGPTLMHeadModel,
            OpenAIGPTTokenizer,
            TransfoXLLMHeadModel,
            TransfoXLTokenizer,
            XLMTokenizer,
            XLMWithLMHeadModel,
            XLNetLMHeadModel,
            XLNetTokenizer, pipeline,
        )
        from .utils import EOS

        MODEL_CLASSES = {
            "gpt2": (GPT2LMHeadModel, GPT2Tokenizer),
            "ctrl": (CTRLLMHeadModel, CTRLTokenizer),
            "openai-gpt": (OpenAIGPTLMHeadModel, OpenAIGPTTokenizer),
            "xlnet": (XLNetLMHeadModel, XLNetTokenizer),
            "transfo-xl": (TransfoXLLMHeadModel, TransfoXLTokenizer),
            "xlm": (XLMWithLMHeadModel, XLMTokenizer),
        }

        model_name_or_path = self._augmenter_arguments.get("model_name_or_path", None)
        stop_token = self._augmenter_arguments.get("stop_token", EOS)
        num_return_sentences = self._augmenter_arguments.get("num_return_sentences", -1)

        assert model_name_or_path is not None, \
            "Please pass a pretrained model name or path to " \
            "DataAugmentation.augmenter_arguments like: " \
            "{'model_name_or_path': 'gpt2'," \
            "......}"
        config = AutoConfig.from_pretrained(model_name_or_path)
        assert config.model_type in MODEL_CLASSES, \
            "Unsupport this model to augment data:{}".format(config.model_type)
        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
        model = AutoModelForCausalLM.from_pretrained(
            model_name_or_path,
            from_tf=bool(".ckpt" in model_name_or_path),
            config=config,
        )
        model.to(self.device)

        text_generator = pipeline(
            "text-generation", model=model, tokenizer=tokenizer,
            device=-1 if self.device == torch.device("cpu") else 0,
        )
        assert os.path.isfile(self._input_dataset), \
            "Please ensure the input dataset is a txt file"
        # map between the prefix and its number of occurrences in the input file
        label2count = dict()
        label2count[0] = 0
        label2count[1] = 0

        lengths = list()

        for sentence in raw_datasets:
            label2count[sentence["label"]] += 1
            length = len(sentence["sentence"].split())
            lengths.append(length)

        from statistics import mean
        m = int(mean(lengths))
        std = int(np.std(lengths)) + 1
        max_length = m + std
        min_length = m - std

        total_count = sum(label2count.values())
        factor = total_count if num_return_sentences <= 0 \
                 else int(math.ceil(num_return_sentences / self._num_samples))
        p0 = label2count[0] / total_count
        p1 = 1 - p0

        prefix_texts = np.random.choice([0, 1], size=(factor,), p=[p0, p1])
        augmented = set()

        with open(self._output_path, 'w', encoding='utf8') as file:
            file.write("label" + "\t" + "sentence" + '\n')
            for prefix in tqdm(prefix_texts):
                loops = 0
                while loops < 2:
                    text_inputs = str(prefix) + '\t'
                    output_sequences = text_generator(
                        text_inputs=text_inputs, early_stopping=True,
                        temperature=self._augmenter_arguments.get("temperature", 1.0),
                        top_k=self._augmenter_arguments.get("k", 0),
                        top_p=self._augmenter_arguments.get("p", 0.9),
                        repetition_penalty=self._augmenter_arguments.get("repetition_penalty", 1.0),
                        do_sample=True,
                        num_return_sequences=self._num_samples,
                        clean_up_tokenization_spaces=True,
                        return_full_text=True,
                        max_length=max_length,
                        min_length=min_length,
                    )
                    l_text_inputs = len(text_inputs.strip())
                    for seq in output_sequences:
                        text = seq['generated_text']
                        text = text[:text.find(stop_token) if stop_token and text.find(
                            stop_token) > -1 else None].strip()
                        text = text[:text.find('\n') if text.find('\n') > -1 else None].strip()
                        if len(text) > l_text_inputs and text not in augmented:
                            file.write(text + '\n')
                            augmented.add(text)
                            loops += 1
                    loops += 1

                    file.flush()

    def mit_data_augmentation(self, extension, raw_datasets):   # pragma: no cover
        """Execute the process of nlpaug data augmentation.

        nlpaug is the library helps you with augmenting nlp for your machine learning projects.
        It provide many augmenter, please refer to: https://github.com/makcedward/nlpaug#augmenter.

        Args:
            extension (str): Suffix name of the original dataset file, now only support "csv" and "json".
            raw_datasets (dict): original datasets, like:label + '\t' + sentence.
                user should set column_names if the sample contains multiple sentences:
                    aug.column_names = ['sentence1', 'sentence2'].
        """
        column_names = raw_datasets.column_names
        if "idx" in column_names:
            column_names.remove("idx")
        args = {} if self._augmenter_arguments is None else self._augmenter_arguments

        aug = get_augmenter_from_type(self._augmenter_type)
        auger = methodcaller(self._augmenter_type, **args)(aug)

        with open(self._output_path, 'w', encoding='utf8') as file:
            if extension == "json":
                writer = json
            else:
                writer = csv.DictWriter(file, fieldnames=column_names, delimiter="\t")
                writer.writeheader()
            for sample in raw_datasets:
                if "idx" in sample:
                    sample.pop("idx")
                if isinstance(self._column_names, list):
                    for column in self._column_names:
                        text = auger.augment(sample[column], n=self._num_samples)
                        sample[column] = text
                else:
                    text = auger.augment(sample[self._column_names], n=self._num_samples)
                    sample[self._column_names] = text
                if extension == "json":
                    text = writer.dumps(sample)
                    file.write(text)
                    file.write("\n")
                else:
                    writer.writerow(sample)
                file.flush()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The MatMulWithBiasAdd Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
from ..onnx_utils import bias_to_int32
import copy


@pattern_registry(pattern_type='MatMulWithBiasAdd')
class MatMulWithBiasAdd(Pattern):
    """The MatMulWithBiasAdd pattern.

    Fuse the original sub-graph into the custom acceleration 'MatMulWithBiasAdd' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'MatMulWithBiasAdd': [
                {
                    'patterns': {
                        'in': [[(0, 'MatMulWithBias'), (1, ['Add', 'AddV2'])]],
                        'out': [[(0, 'MatMulWithBiasAdd')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 1
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            0: [1]
                        }, {
                            0: [2]
                        }, {
                            1: [0, 1]
                        }], [[0, 1, 2, 3], 4]]
                    },
                    'output_tensors': {
                        0: [[{
                            1: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0]
                },
            ]
        }

        def _set_attr(new_node_names, ret_old_nodes, model):
            for i in range(len(new_node_names)):
                mat_node_idx = model.get_node_id(new_node_names[i][0])
                attr = OrderedDict()
                if 'src0_perm' in ret_old_nodes[i][0].attr.keys():
                    attr['src0_perm'] = ret_old_nodes[i][0].attr['src0_perm']
                if 'src1_perm' in ret_old_nodes[i][0].attr.keys():
                    attr['src1_perm'] = ret_old_nodes[i][0].attr['src1_perm']
                attr['append_op'] = 'sum'
                model.nodes[mat_node_idx].attr = attr

        pattern_dict = pattern_mapping_config['MatMulWithBiasAdd'][0]
        model, new_node_names, ret_old_nodes = util.pattern_mapping("MatMulWithBiasAdd",
                                                                    pattern_dict, model)
        if len(new_node_names) != 0:
            _set_attr(new_node_names, ret_old_nodes, model)

            return model

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The RemoveSlice Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
import copy


@pattern_registry(pattern_type='RemoveSlice')
class RemoveSlice(Pattern):
    """The RemoveSlice pattern.

    RemoveSlice
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        if model.framework_modeling_config['framework'] != 'torch':
            return model

        remove_list = []
        first_one = False 
        output_tensor = None
        for node in model.nodes:
            if node.op_type == 'SliceMask':
                if first_one == False:
                    output_tensor = copy.deepcopy(node.output_tensors[0])
                    first_one = True
                    continue
                next_node = model.get_node_by_name(node.output_tensors[0].dest_op[0])
                for i in range(len(next_node.input_tensors)):
                    if next_node.input_tensors[i].name ==  node.output_tensors[0].name:
                        next_node.input_tensors[i] = output_tensor
                        output_tensor.dest_op.append(model.get_node_by_name(node.output_tensors[0].dest_op[0]).name)
                      

                remove_list.append(node.name)
                 
        model.remove_nodes(remove_list)
        return model



import shutil
import numpy as np
import unittest
import tensorflow as tf
from datasets import load_dataset, load_metric
from transformers import (TFAutoModelForSequenceClassification, AutoTokenizer,
                          DefaultDataCollator, HfArgumentParser,
                          TFTrainingArguments, set_seed)
from intel_extension_for_transformers.optimization import (
    AutoDistillationConfig,
    TFDistillationConfig,
    metrics,
)
from intel_extension_for_transformers.optimization.optimizer_tf import TFOptimization
from intel_extension_for_transformers.optimization.utils.utility_tf import distributed_init

def compute_metrics(preds, label_ids):
    metric = load_metric("glue", "sst2")
    preds = preds["logits"]
    preds = np.argmax(preds, axis=1)
    result = metric.compute(predictions=preds, references=label_ids)
    if len(result) > 1:
        result["combined_score"] = np.mean(list(result.values())).item()
    return result

class TestAutoDistillation(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.strategy = tf.distribute.MultiWorkerMirroredStrategy()
        set_seed(42)
        self.model = TFAutoModelForSequenceClassification.from_pretrained(
            'distilbert-base-uncased')
        self.teacher_model = TFAutoModelForSequenceClassification.from_pretrained(
            'distilbert-base-uncased-finetuned-sst-2-english')

        raw_datasets = load_dataset("glue", "sst2")["validation"]
        self.tokenizer = AutoTokenizer.from_pretrained("distilbert-base-uncased")
        non_label_column_names = [
            name for name in raw_datasets.column_names if name != "label"
        ]

        def preprocess_function(examples):
            # Tokenize the texts
            args = ((examples['sentence'], ))
            result = self.tokenizer(*args,
                                    padding=True,
                                    max_length=64,
                                    truncation=True)
            return result

        raw_datasets = raw_datasets.map(preprocess_function,
                                        batched=True,
                                        load_from_cache_file=False)
        data_collator = DefaultDataCollator(return_tensors="tf")
        dataset = raw_datasets.select(range(10))
        self.dummy_dataset = dataset.to_tf_dataset(
            columns=[
                col for col in dataset.column_names
                if col not in set(non_label_column_names + ["label"])
            ],
            shuffle=False,
            batch_size=2,
            collate_fn=data_collator,
            drop_remainder=False,
            # `label_cols` is needed for user-defined losses, such as in this example
            # datasets v2.3.x need "labels", not "label"
            label_cols=["labels"]
            if "label" in dataset.column_names else None,
        )
        parser = HfArgumentParser(TFTrainingArguments)
        self.args = parser.parse_args_into_dataclasses(args=[
            "--output_dir", "./distilled_model",
            "--per_device_eval_batch_size", "2"
        ])[0]
        optimizer = tf.keras.optimizers.Adam(
            learning_rate=self.args.learning_rate,
            beta_1=self.args.adam_beta1,
            beta_2=self.args.adam_beta2,
            epsilon=self.args.adam_epsilon,
            clipnorm=self.args.max_grad_norm,
        )
        loss_fn = tf.keras.losses.SparseCategoricalCrossentropy(
            from_logits=True, reduction=tf.keras.losses.Reduction.SUM)
        metrics = ["accuracy"]
        self.model.compile(optimizer=optimizer, loss=loss_fn, metrics=metrics)

        self.optimizer = TFOptimization(model=self.model,
                                        args=self.args,
                                        train_dataset=self.dummy_dataset,
                                        eval_dataset=self.dummy_dataset,
                                        compute_metrics=compute_metrics)

    @classmethod
    def tearDownClass(self):
        shutil.rmtree('./tmp', ignore_errors=True)
        shutil.rmtree('./distilled_model', ignore_errors=True)

    def test_tf_auto_distillation(self):
        for search_algorithm in ['BO', 'Grid', 'Random']:
            max_trials = 6 if search_algorithm == 'Random' else 3
            autodistillation_config = AutoDistillationConfig(
                search_space={
                    'hidden_size': [120, 240],
                    'intermediate_size': [256, 512]
                },
                search_algorithm=search_algorithm,
                max_trials=max_trials,
                metrics=[
                    metrics.Metric(name="metric", greater_is_better=False)
                ],
                knowledge_transfer=TFDistillationConfig(
                    train_steps=[3],
                    loss_types=['CE', 'CE'],
                    loss_weights=[0.5, 0.5],
                    temperature=1.0
                ),
                regular_distillation=TFDistillationConfig(
                    train_steps=[3],
                    loss_types=['CE', 'CE'],
                    loss_weights=[0.5, 0.5],
                    temperature=1.0
                )
            )
            best_model_archs1 = self.optimizer.autodistill(
                autodistillation_config,
                self.teacher_model,
                model_cls=TFAutoModelForSequenceClassification,
                train_func=None,
                eval_func=None
            )

            best_model_archs2 = self.optimizer.autodistill(
                autodistillation_config,
                self.teacher_model,
                model_cls=TFAutoModelForSequenceClassification,
                train_func=self.optimizer.build_train_func,
                eval_func=self.optimizer.builtin_eval_func
            )
            # check best model architectures
            self.assertTrue(len(best_model_archs1) > 0)
            self.assertTrue(len(best_model_archs2) > 0)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The TorchEmbedding pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util


@pattern_registry(pattern_type='TorchEmbedding')
class TorchEmbedding(Pattern):
    """The TorchEmbedding pattern.

    Fuse the original sub-graph into the custom acceleration 'TorchEmbedding' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        if model.framework_modeling_config['framework'] != 'torch':
            return model

        node_idx = []
        for node in model.nodes:
            if node.op_type == 'Gather':
                if len(node.input_tensors) >=2 and len(node.input_tensors[1].source_op) == 1:
                    view_node = model.get_node_by_name(node.input_tensors[1].source_op[0])
                else:
                    continue
                if len(view_node.input_tensors) >=2 and len(view_node.input_tensors[1].source_op) == 1:
                    shape_node = model.get_node_by_name(view_node.input_tensors[1].source_op[0])
                else:
                    continue
                if shape_node.op_type != 'Shape' or view_node.op_type != 'View':
                    return model
                shape_node.op_type = 'Reshape'
                shape_node.attr.clear()
                shape_node.attr['dst_shape'] = '-1'

                node.input_tensors[1] = node.input_tensors[0]
                shape_node.output_tensors[0].dest_op = [node.name]
                node.input_tensors[0] = shape_node.output_tensors[0]
                node.attr['axis'] = 0
                node.attr['batch_dims'] = 0
                out_tensor = node.output_tensors[0]
                node.output_tensors[0] = view_node.output_tensors[0]
                node.output_tensors[0].source_op = [node.name]

                view_node.op_type = 'Reshape'
                node.output_tensors[0].dest_op = [view_node.name]
                view_node.input_tensors[0] = node.output_tensors[0]
                out_tensor.source_op = [view_node.name]
                view_node.output_tensors[0] = out_tensor
                view_node.input_tensors[1] = shape_node.input_tensors[0]
                view_node.attr.clear()
                view_node.attr['dst_shape'] = '-1,-1,-1'
                view_node.attr['dims'] = '0,1'
                view_node.attr['mul'] = '0,1'

                node_idx.append((model.get_node_id(view_node.name), model.get_node_id(node.name)))
        for idx1, idx2 in node_idx:
            tmp_node = model.nodes[idx1]
            model.nodes[idx1] = model.nodes[idx2]
            model.nodes[idx2] = tmp_node
        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The EmbeddingBag Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util


@pattern_registry(pattern_type='EmbeddingBag')
class EmbeddingBag(Pattern):
    """The EmbeddingBag pattern.

    Fuse the original sub-graph into the custom acceleration 'EmbeddingBag' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'EmbeddingBag': [
                # DLRM onnx model
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (3, 'Unsqueeze'), (4, 'Concat'),
                                (5, 'Slice'), (6, 'Shape'), (7, 'Gather'), (8, 'Loop')],
                               [(), (2, 'Gather'), (4, 'Concat')]],
                        'out': [[(0, 'Gather'), (1, 'Reshape'), (2, 'EmbeddingBag')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 2,
                        1: 6,
                        2: 8,
                    },
                    'input_tensors': {
                        0: [[{
                         2: [1]}
                         , {2: [0]}], [[0,1], 2]],
                        1: [[], [[], 1]],
                        2: [[{
                         0: [0]}
                         , {8: [0]}], [[0,2], 3]],
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2 : [[{8: [0]}], [[0], 1]]
                    },
                    'returns': [2]
                },

                {
                    'patterns': {
                        'in': [[(0, 'Squeeze'), (0, 'EmbeddingBag')]],
                        'out': [[(0, 'Reshape'), (0, 'EmbeddingBag')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0,
                        1: 1
                    },
                    'input_tensors': {
                        0: [[{0: [0]}], [[0], 1]],
                        1: [[{1: [1]}, {1: [2]}], [[1,2], 3]],
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1 : [[{1: [0]}], [[0], 1]]
                    },
                    'returns': []
                },

            ]
        }

        def _set_attr(hidden_size, axis, batch_dims, node_names, model):
            attr1 = OrderedDict()
            attr1['dst_shape'] = -1
            attr2 = OrderedDict()
            attr2['axis'] = axis
            attr2['batch_dims'] = batch_dims

            reshape_0_node_idx = model.get_node_id(node_names[1])
            model.nodes[reshape_0_node_idx].attr = attr1

            gather_node_idx = model.get_node_id(node_names[0])
            model.nodes[gather_node_idx].attr = attr2

        for i in range(len(pattern_mapping_config['EmbeddingBag'])):
            pattern_dict = pattern_mapping_config['EmbeddingBag'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("EmbeddingBag", 
                                                                        pattern_dict, model)

            if len(new_node_names) != 0 and i == 0:
                for j in range(len(new_node_names)):
                    gatherv2_node = ret_old_nodes[j][0]
                    hidden_size = int(gatherv2_node.input_tensors[1].shape[-1])
                    axis = gatherv2_node.attr['axis']
                    batch_dims = gatherv2_node.attr['batch_dims']
                    _set_attr(hidden_size, axis, batch_dims, new_node_names[j], model)
            elif len(new_node_names) != 0 and i == 1:
                for j in range(len(new_node_names)):
                    attr = OrderedDict()
                    attr['dst_shape'] = -1
                    reshape_node_idx = model.get_node_id(new_node_names[j][0])
                    model.nodes[reshape_node_idx].attr = attr


        return model



import os
import onnx
import shutil
import torch
import torch.utils.data as data
import unittest
import transformers
import logging
import numpy as np

from intel_extension_for_transformers.optimization.trainer import NLPTrainer
from intel_extension_for_transformers.optimization.utils.models.modeling_roberta_dynamic import RobertaForQuestionAnswering
from intel_extension_for_transformers.optimization.utils.models.modeling_bert_dynamic import BertForQuestionAnswering
from intel_extension_for_transformers.optimization.dynamic.drop_and_restore_utils import (
    sample_length_configuration,
    sample_layer_configuration
)

from intel_extension_for_transformers.optimization.dynamic.evolution import (
    approx_ratio, inverse, store2str, Evolution
)

transformers.models.roberta.modeling_roberta.RobertaForQuestionAnswering = RobertaForQuestionAnswering
transformers.models.bert.modeling_bert.BertForQuestionAnswering = BertForQuestionAnswering

from transformers import (
    AutoConfig,
    AutoModelForQuestionAnswering,
    AutoTokenizer,
    TrainingArguments,
)

from intel_extension_for_transformers.optimization import (
    DynamicLengthConfig,
)


os.environ["WANDB_DISABLED"] = "true"
os.environ["DISABLE_MLFLOW_INTEGRATION"] = "true"
BERT_MODEL = "bert-base-uncased"
MINILM_MODEL = "sguskin/minilmv2-L6-H384-squad1.1"
MAX_LENGTH = 27
SANDWICHES = 2
LENGTHDROP_RATIO = 0.2
LAYER_DROPOUT_PROB = 0.2
LAYER_DROPOUT_BOUND = 0.2
EVO_ITER = 2
POPULATION_SIZE = 2
MUTATION_SIZE = 2
CROSSOVER_SIZE = 2
NUM_LAYERS = 6


class DummyDataset(data.Dataset):
    def __init__(self, labels=False, type=None):
        MODEL_NAME = BERT_MODEL if type=='bert' else MINILM_MODEL
        self.tokenizer = AutoTokenizer.from_pretrained(MODEL_NAME)
        self.sequence_a = "intel-extension-for-transformers is based in SH"
        self.sequence_b = "Where is intel-extension-for-transformers based?"
        self.encoded_dict = self.tokenizer(self.sequence_a, self.sequence_b, padding="max_length", max_length=MAX_LENGTH)
        if labels:
            self.encoded_dict['start_positions'] = [21]
            self.encoded_dict['end_positions'] = [25]
        
    def __len__(self):
        return 1

    def __getitem__(self, index):
        """Returns one data pair (source and target)."""
        return self.encoded_dict


def check_onnx(model_path, dataloader):
    import onnxruntime as ort
    import numpy as np
    # Check onnxruntime
    ort_session = ort.InferenceSession(model_path)
    # Preprocess input for onnxruntime
    it = iter(dataloader)
    input = next(it)
    input_names = list(input.keys())
    for k in input_names:
        if 'label' in k:
            input.pop(k)
        else:
            input[k] = np.array(input[k])
    # Run onnxruntime inference session
    ort_session.run(None, input,)
    return True

class TestDynamicLengthInferenceRoberta(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.model = AutoModelForQuestionAnswering.from_pretrained(
            MINILM_MODEL
        )
        self.dummy_dataset = DummyDataset(type='roberta')
        self.dynamic_trainer = NLPTrainer(
            model=self.model,
            train_dataset=self.dummy_dataset,
            eval_dataset=self.dummy_dataset,
        )

    @classmethod
    def tearDownClass(self):
        shutil.rmtree('./mlruns', ignore_errors=True)
        shutil.rmtree('./tmp_trainer', ignore_errors=True)
        shutil.rmtree('./dynamic-model.onnx', ignore_errors=True)

    def test_dynamic_inference(self):
        full_output = self.dynamic_trainer.predict(self.dummy_dataset).predictions
        
        dynamic_length_config = DynamicLengthConfig(
            const_rate=0.2,
            max_length=MAX_LENGTH
        )
        self.dynamic_trainer.set_dynamic_config(dynamic_length_config)
        
        dynamic_output = self.dynamic_trainer.predict(self.dummy_dataset).predictions
          
        self.assertTrue((full_output[0] != dynamic_output[0]).any())
  
        #check onnx
        self.dynamic_trainer.export_to_onnx('dynamic-model.onnx')
        self.assertTrue(check_onnx('dynamic-model.onnx', self.dynamic_trainer.get_eval_dataloader()))

class TestDynamicLengthInferenceBert(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.model = AutoModelForQuestionAnswering.from_pretrained(
            BERT_MODEL
        )
        self.dummy_dataset = DummyDataset(type='bert')
        self.dynamic_trainer = NLPTrainer(
            model=self.model,
            train_dataset=self.dummy_dataset,
            eval_dataset=self.dummy_dataset,
        )

    @classmethod
    def tearDownClass(self):
        shutil.rmtree('./mlruns', ignore_errors=True)
        shutil.rmtree('./tmp_trainer', ignore_errors=True)
        shutil.rmtree('./dynamic-model.onnx', ignore_errors=True)

    def test_dynamic_inference(self):
        full_output = self.dynamic_trainer.predict(self.dummy_dataset).predictions
        
        dynamic_length_config = DynamicLengthConfig(
            const_rate=0.2,
            max_length=MAX_LENGTH
        )
        self.dynamic_trainer.set_dynamic_config(dynamic_length_config)
        
        dynamic_output = self.dynamic_trainer.predict(self.dummy_dataset).predictions
          
        self.assertTrue((full_output[0] != dynamic_output[0]).any())
  
        #check onnx
        self.dynamic_trainer.export_to_onnx('dynamic-model.onnx')
        self.assertTrue(check_onnx('dynamic-model.onnx', self.dynamic_trainer.get_eval_dataloader()))



class TestDynamicLengthTrainingRoberta(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.model = AutoModelForQuestionAnswering.from_pretrained(
            MINILM_MODEL
        )
        self.dummy_dataset = DummyDataset(labels=True,type='roberta')
        training_args = TrainingArguments(
            output_dir='./results',
            num_train_epochs=1,
        )
        self.dynamic_trainer = NLPTrainer(
            args=training_args,
            model=self.model,
            train_dataset=self.dummy_dataset,
            eval_dataset=self.dummy_dataset,
        )

    @classmethod
    def tearDownClass(self):
        shutil.rmtree('./mlruns', ignore_errors=True)
        shutil.rmtree('./tmp_trainer', ignore_errors=True)
        shutil.rmtree('./results', ignore_errors=True)


    def test_dynamic_training(self):
        dynamic_length_config = DynamicLengthConfig(
            dynamic_training=True,
            num_sandwich=SANDWICHES,
            length_drop_ratio_bound=LENGTHDROP_RATIO,
            layer_dropout_prob=LAYER_DROPOUT_PROB,
            layer_dropout_bound=LAYER_DROPOUT_BOUND,
            max_length=MAX_LENGTH
        )

        self.dynamic_trainer.set_dynamic_config(dynamic_config=dynamic_length_config)
        self.dynamic_trainer.train()

class TestDynamicLengthTrainingBert(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.model = AutoModelForQuestionAnswering.from_pretrained(
            BERT_MODEL
        )
        self.dummy_dataset = DummyDataset(labels=True,type='bert')
        training_args = TrainingArguments(
            output_dir='./results',
            num_train_epochs=1,
        )
        self.dynamic_trainer = NLPTrainer(
            args=training_args,
            model=self.model,
            train_dataset=self.dummy_dataset,
            eval_dataset=self.dummy_dataset,
        )

    @classmethod
    def tearDownClass(self):
        shutil.rmtree('./mlruns', ignore_errors=True)
        shutil.rmtree('./tmp_trainer', ignore_errors=True)
        shutil.rmtree('./results', ignore_errors=True)


    def test_dynamic_training(self):
        dynamic_length_config = DynamicLengthConfig(
            dynamic_training=True,
            num_sandwich=SANDWICHES,
            length_drop_ratio_bound=LENGTHDROP_RATIO,
            layer_dropout_prob=LAYER_DROPOUT_PROB,
            layer_dropout_bound=LAYER_DROPOUT_BOUND,
            max_length=MAX_LENGTH
        )

        self.dynamic_trainer.set_dynamic_config(dynamic_config=dynamic_length_config)
        self.dynamic_trainer.train()


class TestEvolutionarySearchRoberta(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.model = AutoModelForQuestionAnswering.from_pretrained(
            MINILM_MODEL
        )
        self.dummy_dataset = DummyDataset(labels=True,type='roberta')
        training_args = TrainingArguments(
            output_dir='./results',
        )

        self.dynamic_trainer = NLPTrainer(
            args=training_args,
            model=self.model,
            train_dataset=self.dummy_dataset,
            eval_dataset=self.dummy_dataset,
        )



    @classmethod
    def tearDownClass(self):
        shutil.rmtree('./mlruns', ignore_errors=True)
        shutil.rmtree('./tmp_trainer', ignore_errors=True)
        shutil.rmtree('./results', ignore_errors=True)


    def test_dynamic_training(self):
        
        dynamic_length_config = DynamicLengthConfig(
            evo_iter=EVO_ITER,
            population_size=POPULATION_SIZE,
            mutation_size=MUTATION_SIZE,
            crossover_size=CROSSOVER_SIZE,
            max_length=MAX_LENGTH,
            evo_eval_metric='eval_loss' # dummy metric just for testing
        )

        self.dynamic_trainer.set_dynamic_config(dynamic_config=dynamic_length_config)
        self.dynamic_trainer.run_evolutionary_search() 

class TestEvolutionarySearchBert(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.model = AutoModelForQuestionAnswering.from_pretrained(
            BERT_MODEL
        )
        self.dummy_dataset = DummyDataset(labels=True,type='bert')
        training_args = TrainingArguments(
            output_dir='./results',
        )

        self.dynamic_trainer = NLPTrainer(
            args=training_args,
            model=self.model,
            train_dataset=self.dummy_dataset,
            eval_dataset=self.dummy_dataset,
        )



    @classmethod
    def tearDownClass(self):
        shutil.rmtree('./mlruns', ignore_errors=True)
        shutil.rmtree('./tmp_trainer', ignore_errors=True)
        shutil.rmtree('./results', ignore_errors=True)


    def test_search(self):
        
        dynamic_length_config = DynamicLengthConfig(
            evo_iter=EVO_ITER,
            population_size=POPULATION_SIZE,
            mutation_size=MUTATION_SIZE,
            crossover_size=CROSSOVER_SIZE,
            max_length=MAX_LENGTH,
            evo_eval_metric='eval_loss'
        )

        self.dynamic_trainer.set_dynamic_config(dynamic_config=dynamic_length_config)
        # self.dynamic_trainer.run_evolutionary_search() # dummy metric just for testing

    def test_search_functions(self):
        dummy_ratios = [inverse(r) for r in np.linspace(approx_ratio(0.2,n=NUM_LAYERS,l=MAX_LENGTH), 1, POPULATION_SIZE + 2)[1:-1]]
        self.assertTrue( len(dummy_ratios) == POPULATION_SIZE, msg='{0},{1}'.format(len(dummy_ratios), POPULATION_SIZE))
        self.assertTrue( all(i < 1 for i in dummy_ratios), msg='{0}'.format(dummy_ratios))
        res = store2str(gene=(MAX_LENGTH,), macs=1234, score=88, method='dummy')
        subs = ('MACs','score','method')
        self.assertTrue( all(i in res for i in subs), msg='{0}'.format(res))
        evo = Evolution(self.model,MAX_LENGTH, 'cpu', None, eval_metric='eval_loss')
        


class TestSampleConfiguration(unittest.TestCase):
   
    def test_sample_length_config(self):
 
        no_drop_lc = tuple( MAX_LENGTH for _ in range(NUM_LAYERS))
        lc = sample_length_configuration(MAX_LENGTH, NUM_LAYERS)

        self.assertTrue( lc == no_drop_lc, msg='{0}, {1}'.format(lc, no_drop_lc))

        rate = 0.1
        next_length = MAX_LENGTH
        lc_const_drop = ()
        for _ in range(NUM_LAYERS):
            next_length = int(np.ceil(next_length * (1 - rate)))
            lc_const_drop += (next_length,)

        lc = sample_length_configuration(MAX_LENGTH, NUM_LAYERS, length_drop_ratio=rate)
        self.assertTrue( lc == lc_const_drop, msg='{0}, {1}'.format(lc, lc_const_drop))
        lc = sample_length_configuration(MAX_LENGTH, NUM_LAYERS, length_drop_prob=0.2)
        self.assertTrue( lc[-1] < MAX_LENGTH , msg='{0}, {1}'.format(lc[-1], MAX_LENGTH))
        lc = sample_length_configuration(MAX_LENGTH, NUM_LAYERS, length_drop_ratio_bound=0.2)
        self.assertTrue( lc[-1] < MAX_LENGTH , msg='{0}, {1}'.format(lc[-1], MAX_LENGTH))

        no_drop_lc = tuple( l for l in range(NUM_LAYERS))
        lc = sample_layer_configuration(NUM_LAYERS, layer_dropout=0)
        self.assertTrue( lc == no_drop_lc, msg='{0}, {1}'.format(lc, no_drop_lc))
        drop_rate = NUM_LAYERS-3
        lc = sample_layer_configuration(NUM_LAYERS, layer_dropout=drop_rate)
        expected = tuple(i for i in range(drop_rate))
        self.assertTrue( lc == expected, msg='{0}, {1}'.format(lc, expected))
        layer_conf = sample_layer_configuration(NUM_LAYERS, layer_dropout_prob=0.2)
        length_conf = sample_length_configuration(MAX_LENGTH, NUM_LAYERS, layer_config=layer_conf, length_drop_ratio=0.1)
        self.assertTrue( len(length_conf) == NUM_LAYERS , msg='{0}, {1}'.format(length_conf, layer_conf))


       

        




if __name__ == "__main__":
    unittest.main()
   


"""Pruning."""
# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

from .utils import process_config, parse_to_prune, \
    check_config, update_params
from .pruners import get_pruner
from .logger import logger

from intel_extension_for_transformers.optimization.utils.utility import LazyImport

LazyImport('torch.nn')
torch = LazyImport('torch')


class Pruning:
    """Pruning.

    The main class to do pruning; it contains at least one Pruner object.

    Args:
        config: a string representing the path to a config file. For config file template, please refer to
            https://github.com/intel/neural-compressor/tree/master/examples/pytorch/nlp/huggingface_models/text-classification/pruning/pytorch_pruner/eager/

    Attributes:
        model: The model object to prune.
        config_file_path: A string representing the path to a config file.
        pruners: A list. A list of Pruner objects.
        pruner_info: A config dict object that contains pruners' information.
    """

    def __init__(self, config):
        """Initialize."""
        self._model = None
        self.pruners = []
        self.pruners_info = process_config(config)

    @property
    def model(self):
        """Obtain model in neural_compressor.model."""
        return self._model

    @model.setter
    def model(self, user_model):
        self._model = user_model

    def update_config(self, *args, **kwargs):
        """Add user-defined arguments to the original configurations.

        The original config of pruning is read from a file.
        However, users can still modify configurations by passing key-value arguments in this function.
        Please note that the key-value arguments' keys could be processed in current configuration.
        """
        for item in self.pruners_info:
            for key in kwargs:
                if key in item.keys():
                    item[key] = kwargs[key]

            update_params(item)
            check_config(item)

    def get_sparsity_ratio(self):
        """Calculate sparsity ratio of a module/layer.

        Returns:
            Three floats.
            elementwise_over_matmul_gemm_conv refers to zero elements' ratio in pruning layers.
            elementwise_over_all refers to zero elements' ratio in all layers in the model.
            blockwise_over_matmul_gemm_conv refers to all-zero blocks' ratio in pruning layers.
        """
        pattern_sparsity_cnt = 0
        element_sparsity_cnt = 0
        for pruner in self.pruners:
            modules = pruner.modules
            sparsity_ratio = pruner.pattern.get_sparsity_ratio(pruner.masks)
            cnt = 0
            for key in modules.keys():
                cnt += modules[key].weight.numel()
            pattern_sparsity_cnt += int(cnt * sparsity_ratio)
            for key in pruner.masks.keys():
                element_sparsity_cnt += torch.sum(pruner.masks[key] == 0).data.item()

        linear_conv_cnt = 0
        param_cnt = 0
        for name, module in self._model.named_modules():
            if type(module).__name__ in ["Linear"] or re.search(r'Conv.d', type(module).__name__) != None:
                linear_conv_cnt += module.weight.numel()

        for n, param in self._model.named_parameters():
            param_cnt += param.numel()
        if linear_conv_cnt == 0:
            blockwise_over_matmul_gemm_conv = 0
            elementwise_over_matmul_gemm_conv = 0
        else:
            blockwise_over_matmul_gemm_conv = float(pattern_sparsity_cnt) / linear_conv_cnt
            elementwise_over_matmul_gemm_conv = float(element_sparsity_cnt) / linear_conv_cnt
        if param_cnt == 0:
            elementwise_over_all = 0
        else:
            elementwise_over_all = float(
                element_sparsity_cnt) / param_cnt

        return elementwise_over_matmul_gemm_conv, elementwise_over_all, blockwise_over_matmul_gemm_conv

    def _generate_pruners(self):
        """Obtain Pruner objects."""
        assert isinstance(self._model, torch.nn.Module)

        for info in self.pruners_info:
            modules = parse_to_prune(info, self._model)
            if modules == {}:
                logger.warning("one pruner hooks no layers, please have a check")

            self.pruners.append(get_pruner(info, modules))
            info['modules'] = [key for key in modules.keys()]
            info['len_of_modules'] = len(info['modules'])
            logger.info(info)

    def on_train_begin(self):
        """Implement at the beginning of training process.

        Before training, ensure that pruners are generated.
        """
        self._generate_pruners()  ##TODO is there better place to place

    def on_epoch_begin(self, epoch):
        """Implement at the beginning of every epoch."""
        for pruner in self.pruners:
            pruner.on_epoch_begin(epoch)

    def on_step_begin(self, local_step):
        """Implement at the beginning of every step."""
        for pruner in self.pruners:
            pruner.on_step_begin(local_step)

    def on_before_optimizer_step(self):
        """Implement before optimizer.step()."""
        for pruner in self.pruners:
            pruner.on_before_optimizer_step()

    def on_step_end(self):
        """Implement at the end of every step."""
        for pruner in self.pruners:
            pruner.on_step_end()

    def on_epoch_end(self):
        """Implement the end of every epoch."""
        for pruner in self.pruners:
            pruner.on_epoch_end()

    def on_train_end(self):
        """Implement the end of training phase."""
        for pruner in self.pruners:
            pruner.on_train_end()
        self.get_sparsity_ratio()

    def on_before_eval(self):
        """Implement at the beginning of evaluation phase."""
        for pruner in self.pruners:
            pruner.on_before_eval()

    def on_after_eval(self):
        """Implement at the end of evaluation phase."""
        for pruner in self.pruners:
            pruner.on_after_eval()

    def on_after_optimizer_step(self):
        """Implement after optimizer.step()."""
        for pruner in self.pruners:
            pruner.on_after_optimizer_step()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine operator mapping file."""

from .op import Operator, operator_registry
from .tensor import Tensor
from ..graph_utils import list2str


# tf.math.reduce_mean(input_tensor, axis=None, keepdims=False, name=None)
# If axis is None, all dimensions are reduced, and a tensor with a single element is returned.
@operator_registry(operator_type='ReduceMean')
class ReduceMean(Operator):
    """Parse the ReduceMean operator to the neural engine."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def set_attr(self, framework, node):
        """Extract the node attr from tensorflow."""
        if framework == 'tensorflow':
            self._attr['axis'] = self._input_tensors[1].data[0]
            self._attr['keep_dims'] = node.attr['keep_dims'].b
        """Extract the node attr from onnxruntime."""
        if framework == 'onnxruntime':
            if len(node.attribute) == 2:
                axis = node.attribute[1].ints
                self._attr['keep_dims'] = bool(node.attribute[0].i)
            if len(node.attribute) == 1:
               axis = node.attribute[0].ints

            if len(axis) == 1:
                self._attr['axis'] = axis[0]
            else:
                self._attr['axis'] = list2str(axis)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine operator mapping file."""

from .op import Operator, operator_registry, list2str, parseTorchListConstruct
from .tensor import Tensor


@operator_registry(operator_type='LayerNormalization')
class LayerNormalization(Operator):
    """Parse the LayerNormalization operator to the neural engine."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def set_attr(self, framework, node):
        """Extract the node attr from frameworks."""
        self._op_type = 'LayerNorm'
        if framework == "onnxruntime":
            axis = node.attribute[1].i
            if axis != -1:
                self._attr['axis'] = axis
            self._attr['epsilon'] = node.attribute[2].f

# Fused_op Mean, AddV2, Mul, etc.
# This pattern has several ops combinations, so the input_tensors and output_tensors may various
@operator_registry(operator_type='LayerNorm')
class LayerNorm(Operator):
    """Register the LayerNorm operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def set_attr(self, framework, node):
        if framework == 'torch':
            """Extract the node attr from torchscript."""
            if node.inputsSize() > 4:
                self._attr['epsilon'] = node.inputsAt(4).toIValue()
                self._attr['normalized_shape'] = list2str(parseTorchListConstruct(node.inputsAt(1)))


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The RemoveRange Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
from .subgraph_matcher import EXECUTOR_TYPE
import numpy as np
import copy

@pattern_registry(pattern_type='RemoveRange')
class RemoveRange(Pattern):
    """The RemoveRange pattern.

    Fuse the original sub-graph into the custom acceleration 'RemoveRange' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """

    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'RemoveRange': [
                {
                    'patterns': {
                        'in': [[(0, 'Range'), (1, 'Div'), (2, 'Pow'), (3, 'Div'), (4, 'Reshape'), (7, 'Matmul')],
                               [(), (5, 'Range'), (6, 'Reshape'), (7, 'Matmul')]
                        ],
                    },
                },
                
                {
                    'patterns': {
                        'in': [[(0, 'Cos'), (1, 'Reshape'), (2, 'Gather'), (3, 'Reshape'), 
                                (4, 'Slice')]
                        ],
                    },
                },
                {
                    'patterns': {
                        'in': [[(0, 'Sin'), (1, 'Reshape'), (2, 'Gather'), (3, 'Reshape'), 
                                (4, 'Slice')]
                        ],
                    },
                },   
                
            ],
        }

        if model.framework_modeling_config['framework'] != 'torch':
            return model

        pattern = pattern_mapping_config['RemoveRange'][0]['patterns']['in']
        patterns_nodes_name = util.search_pattern(pattern, model)
        once_flag = False
        for pattern_nodes_name in patterns_nodes_name:
            if once_flag == False:
                once_flag = True
                continue
            remove_list = []
            for node in pattern_nodes_name:
                if not (isinstance(node, list)):
                    remove_list.append(node)
            model.remove_nodes(remove_list)
            
        pattern = pattern_mapping_config['RemoveRange'][1]['patterns']['in']
        patterns_nodes_name = util.search_pattern(pattern, model)
        once_flag2 = False
        
        keep_list = []
        for pattern_nodes_name in patterns_nodes_name:
            if once_flag2 == False :
                once_flag2 = True
                keep_list = pattern_nodes_name
                continue
            remove = model.get_node_by_name(pattern_nodes_name[-2])
            keep = model.get_node_by_name(keep_list[-2])
            
            next_node = model.get_node_by_name(remove.output_tensors[0].dest_op[0])
            keep.output_tensors[0].dest_op.append(next_node.name)
            next_node.input_tensors[1] = keep.output_tensors[0]
            for node_name in pattern_nodes_name:
                if node_name not in keep_list and not (isinstance(node_name, list)):
                    model.remove_nodes([node_name])
            
        pattern = pattern_mapping_config['RemoveRange'][2]['patterns']['in']
        patterns_nodes_name = util.search_pattern(pattern, model)
        once_flag2 = False
        
        keep_list = []
        for pattern_nodes_name in patterns_nodes_name:
            if once_flag2 == False :
                once_flag2 = True
                keep_list = pattern_nodes_name
                continue
            remove = model.get_node_by_name(pattern_nodes_name[-2])
            keep = model.get_node_by_name(keep_list[-2])
            
            next_node = model.get_node_by_name(remove.output_tensors[0].dest_op[0])
            keep.output_tensors[0].dest_op.append(next_node.name)
            next_node.input_tensors[1] = keep.output_tensors[0]
            for node_name in pattern_nodes_name:
                if node_name not in keep_list and not (isinstance(node_name, list)):
                    model.remove_nodes([node_name])
            
        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The Transformer2Dmodel_ConstantOfShapeWithMul Pattern."""

from .pattern import Pattern, pattern_registry
from .. import graph_utils as util
from .. import logger


@pattern_registry(pattern_type='Transformer2Dmodel_ConstantOfShapeWithMul')
class Transformer2Dmodel_ConstantOfShapeWithMul(Pattern):
    """The Transformer2Dmodel_ConstantOfShapeWithMul pattern.

    Fuse the original sub-graph into the custom acceleration 'Transformer2Dmodel_ConstantOfShapeWithMul' graph.
    The search strategy is based on the following pattern mapping configs for the stable diffusionV1-5.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'Transformer2Dmodel_ConstantOfShapeWithMul': [
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'), (9, 'Concat'),
                                (10, 'ConstantOfShape'), (11, 'Mul'), (12, 'Add'), (13, 'Softmax'),
                                (14, 'Cast')],
                               [(), (3, 'Shape'), (4, 'Gather'), (5, 'Unsqueeze'), (9, 'Concat')],
                               [(), (6, 'Shape'), (7, 'Gather'), (8, 'Unsqueeze'), (9, 'Concat')]],
                    },
                },
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'), (9, 'Concat'),
                                (10, 'ConstantOfShape'), (11, 'Mul'), (12, 'Add'), (13, 'Softmax'),
                                (14, 'Cast')],
                               [(), (3, 'Shape'), (4, 'Gather'), (5, 'Unsqueeze'), (9, 'Concat')],
                               [(), (6, 'Shape'), (7, 'Gather'), (8, 'Unsqueeze'), (9, 'Concat')]],
                        'out': [[(0, 'Softmax')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 13,
                    },
                    'input_tensors': {
                        0: [[{
                            12: [0]
                        }], [[0], 1]],
                    },
                    'output_tensors': {
                        0: [[{
                            14: [0]
                        }], [[0], 1]],
                    },
                    'returns': [9, 11, 13]
                },
            ]
        }

        pattern = pattern_mapping_config['Transformer2Dmodel_ConstantOfShapeWithMul'][0]['patterns']['in']
        patterns_nodes_name = util.search_pattern(pattern, model)
        mul = -1
        if len(patterns_nodes_name) != 0:
            for j in range(len(patterns_nodes_name)):
                mul_idx = model.get_node_id(patterns_nodes_name[j][11])
                if int(model.nodes[mul_idx].input_tensors[1].data) == 0:
                    mul = 0

        if mul == 0:
            pattern_dict = pattern_mapping_config['Transformer2Dmodel_ConstantOfShapeWithMul'][1]
            model, new_node_names, ret_old_nodes = util.pattern_mapping(
                "Transformer2Dmodel_ConstantOfShapeWithMul", pattern_dict, model)
            if len(new_node_names) != 0:
                logger.info('TextEncoder_AttentionMaskAddReshape mathched...')
                logger.debug('TextEncoder_AttentionMaskAddReshape = {}'.format(new_node_names))

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The Transformer2Dmodel_FFNInputSlice pattern."""

from .pattern import Pattern, pattern_registry
from collections import OrderedDict
from .. import graph_utils as util
from .. import logger


@pattern_registry(pattern_type='Transformer2Dmodel_FFNInputSlice')
class Transformer2Dmodel_FFNInputSlice(Pattern):
    """The Transformer2Dmodel_FFNInputSlice pattern.

    Fuse the original sub-graph into the custom acceleration 'Transformer2Dmodel_FFNInputSlice' graph.
    The search strategy is based on the following pattern mapping configs for the stable textEncoderV1-5.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            # must be placed ahead of the FFNSlice_1 pattern
            'Transformer2Dmodel_FFNInputSlice': [
                {
                    'patterns': {
                        'in': [[(0, 'MatMulWithBias'), (1, 'Shape'), (2, 'Gather'), (3, 'Add'), (4, 'Div'),
                                (5, 'Mul'), (6, 'Slice')]],
                    },
                },
            ]
        }
        #
        pattern = pattern_mapping_config['Transformer2Dmodel_FFNInputSlice'][0]['patterns']['in']
        patterns_nodes_name = util.search_pattern(pattern, model)
        logger.info('Transformer2Dmodel_FFNInputSlice mathched...')
        logger.debug('Transformer2Dmodel_FFNInputSlice = {}'.format(patterns_nodes_name))
        if len(patterns_nodes_name) != 0:
            for j in range(len(patterns_nodes_name)):
                matmul_node = model.get_node_by_name(patterns_nodes_name[j][0])
                bias_shape = matmul_node.input_tensors[1].data.shape[1]

                slice_node = model.get_node_by_name(patterns_nodes_name[j][6])
                if 'starts' in slice_node.attr and slice_node.attr['starts'] == '0':
                    slice_node.input_tensors.pop(1)
                    attr = OrderedDict()
                    attr['starts'] = 0
                    attr['ends'] = int(bias_shape / 2)
                    attr['axes'] = 1
                    attr['steps'] = 1
                    slice_node.attr = attr
                else:
                    attr = OrderedDict()
                    attr['starts'] = int(bias_shape / 2)
                    attr['ends'] = bias_shape
                    attr['axes'] = 1
                    attr['steps'] = 1
                    slice_node.attr = attr

        return model


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The InsertQuantNode Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
from ..ops import Tensor
from .subgraph_matcher import EXECUTOR_TYPE
import numpy as np


@pattern_registry(pattern_type='InsertQuantNode')
class InsertQuantNode(Pattern):
    """The InsertQuantNode pattern.

    Fuse the original sub-graph into the custom acceleration 'InsertQuantNode' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """

    def __call__(self, model):
        """The __call__ function of this pattern class."""
        def get_scale_zp(tensor_min_data, tensor_max_data, dtype):
            # for asym quant
            if (dtype == 'u8'):
                max_sub_min = tensor_max_data - tensor_min_data
                scale = np.where(max_sub_min == 0., 0., 255. / max_sub_min)
                zero_point = -tensor_min_data * scale
            # for sym quant
            elif (dtype == 's8'):
                max_abs = np.maximum(np.fabs(tensor_min_data), np.fabs(tensor_max_data))
                scale = np.where(max_abs == 0., 0., 127. / max_abs)
                zero_point = 0.
            return scale, zero_point

        quant_info = util.get_quant_info()
        if not quant_info:
            return model

        unique_quant_nodes = {}
        node_idx = 0
        while node_idx < len(model.nodes):
            node = model.nodes[node_idx]
            if node.op_type in EXECUTOR_TYPE and \
              (EXECUTOR_TYPE[node.op_type] == "InnerProduct" or \
                EXECUTOR_TYPE[node.op_type] == "Matmul"):
                input_size = len(node.input_tensors)
                if "reshape_dims" in node.attr:
                    input_size -= 1
                insert_offset = input_size

                idx = 0
                while idx < input_size:
                    input_tensor = node.input_tensors[idx]
                    input_name = input_tensor.name
                    if input_name in quant_info and idx < 3:
                        quant_min = Tensor(
                            name=input_name + "_min",
                            shape=[quant_info[input_name][0].size],
                            data=np.array(quant_info[input_name][0].astype("float32")),
                            dtype="fp32")
                        quant_max = Tensor(
                            name=input_name + "_max",
                            shape=[quant_info[input_name][1].size],
                            data=np.array(quant_info[input_name][1].astype("float32")),
                            dtype="fp32")

                        if "insert" in quant_info[input_name][2]:
                            if input_name in unique_quant_nodes:
                                quantize_op_name = unique_quant_nodes[input_name]
                                quant_node = model.get_node_by_name(quantize_op_name)
                                quant_node.output_tensors[0].dest_op.append(node.name)
                                node.input_tensors[idx] = quant_node.output_tensors[0]
                            else:
                                quant_dtype = "u8" if "u8" in quant_info[input_name][2] else "s8"
                                quant_dtype = "s8" if EXECUTOR_TYPE[
                                    node.op_type] == "Matmul" else quant_dtype
                                quant_output = Tensor(name=input_name + "_quant",
                                                    source_op=[node.name + "_quant_" + str(idx)],
                                                    dest_op=[node.name],
                                                    dtype=quant_dtype)
                                quantize_op = util.construct_node(
                                    node_name=node.name + "_quant_" + str(idx),
                                    op_type='Quantize',
                                    input_tensors=[input_tensor, quant_min, quant_max],
                                    output_tensors=[quant_output],
                                    attr=OrderedDict({'output_dtype': quant_dtype}))
                                unique_quant_nodes[input_name] = quantize_op.name
                                node.input_tensors[idx] = quant_output
                                insert_idx = model.get_node_id(node.name)
                                model.insert_nodes(insert_idx, [quantize_op])
                                node_idx += 1
                            # insert src0/src1 min and max tensor
                            model.change_node_input_tensors(node.name, insert_offset + 2 * idx + 0,
                                                            quant_min, 'insert')
                            model.change_node_input_tensors(node.name, insert_offset + 2 * idx + 1,
                                                            quant_max, 'insert')
                        if "weight" in quant_info[input_name][2]:
                            # insert weight min and max tensor
                            model.change_node_input_tensors(node.name, insert_offset + 2 * idx + 0,
                                                            quant_min, 'insert')
                            model.change_node_input_tensors(node.name, insert_offset + 2 * idx + 1,
                                                            quant_max, 'insert')
                        if "output" in quant_info[input_name][2]:
                            output_name = node.output_tensors[0].name
                            quant_min = Tensor(
                                name=output_name + "_min",
                                shape=[quant_info[input_name][3].size],
                                data=np.array(quant_info[input_name][3].astype("float32")),
                                dtype="fp32")
                            quant_max = Tensor(
                                name=output_name + "_max",
                                shape=[quant_info[input_name][4].size],
                                data=np.array(quant_info[input_name][4].astype("float32")), 
                                dtype="fp32")
                            # insert output min and max tensor
                            model.change_node_input_tensors(node.name, insert_offset + 4,
                                                            quant_min, 'insert')
                            model.change_node_input_tensors(node.name, insert_offset + 5,
                                                            quant_max, 'insert')
                            util.insert_quant_info(node.name, [])
                    idx += 1
            node_idx += 1
        # remove fall back quant nodes
        remove_list=[]
        for node in model.nodes:
            if node.op_type in EXECUTOR_TYPE and \
              (EXECUTOR_TYPE[node.op_type] == "InnerProduct" or \
                EXECUTOR_TYPE[node.op_type] == "Matmul"):
                src0_dtype = node.input_tensors[0].dtype == "u8" or \
                             node.input_tensors[0].dtype == "s8"
                src1_dtype = node.input_tensors[1].dtype == "u8" or \
                             node.input_tensors[1].dtype == "s8"
                if src0_dtype ^ src1_dtype:
                    src0_source_op = node.input_tensors[0].source_op
                    src1_source_op = node.input_tensors[1].source_op
                    if src0_dtype:
                        remove_list.append(src0_source_op[0])
                        node.input_tensors[0] = \
                        model.get_node_by_name(src0_source_op[0]).input_tensors[0]
                        node.input_tensors[0].dest_op = [node.name]
                    else: 
                        remove_list.append(src1_source_op[0])
                        node.input_tensors[1] = \
                        model.get_node_by_name(src1_source_op[0]).input_tensors[0]
                        node.input_tensors[1].dest_op = [node.name]
                    remove_tensors_list = []
                    for idx, input_name in enumerate(node.input_tensors):
                        if "_min" in input_name.name or "_max" in input_name.name:
                            remove_tensors_list.append(idx)
                    for remove_idx in remove_tensors_list:
                        model.change_node_input_tensors(node.name, remove_tensors_list[0], None,
                                                    'remove')
        model.remove_nodes(remove_list)

        # remove duplicate tensors
        for node in model.nodes:
            if node.name in quant_info:
                input_tensor_set = set()
                remove_tensors_list = []
                for idx, input_tensor in enumerate(node.input_tensors):
                    input_name = input_tensor.name
                    sz = len(input_tensor_set)
                    input_tensor_set.add(input_name)
                    new_sz = len(input_tensor_set)
                    if new_sz == sz:
                        remove_tensors_list.append(idx)
                for remove_idx in remove_tensors_list:
                    model.change_node_input_tensors(node.name, remove_tensors_list[0], None,
                                                    'remove')

        # Bias compensation for inner product fp32 bias to int32
        for node in model.nodes:
            if node.op_type in EXECUTOR_TYPE and EXECUTOR_TYPE[
                    node.op_type] == "InnerProduct" and len(node.input_tensors) > 4:
                bias_fp32 = node.input_tensors[2].data
                weight_s8 = node.input_tensors[1].data
                offset = 0
                if 'append_op' in node.attr and node.attr['append_op'] in ['binary_add', 'sum']:
                    offset = 1
                input_data_min = node.input_tensors[offset + 3].data
                dtype = node.input_tensors[0].dtype
                input_scale, input_zero_point = get_scale_zp(node.input_tensors[offset + 3].data,
                                                             node.input_tensors[offset + 4].data,
                                                             dtype)
                weight_scale, weight_zero_point = get_scale_zp(node.input_tensors[offset + 5].data,
                                                               node.input_tensors[offset + 6].data,
                                                               's8')
                if dtype == "u8":
                    if "src1_perm" in node.attr and node.attr["src1_perm"] == '1,0':
                        bias_zero_point = input_scale * input_data_min * \
                            np.sum(weight_s8.astype(float), axis=0)
                    else:
                        bias_zero_point = input_scale * input_data_min * \
                            np.sum(weight_s8.astype(float), axis=-1)
                    bias_s32 = bias_fp32 * input_scale * weight_scale
                    bias_s32 = np.round(bias_s32 + bias_zero_point).astype(np.int32)
                    node.input_tensors[2].data = bias_s32
                    node.input_tensors[2].dtype = 's32'
                else:
                    bias_s32 = np.round(bias_fp32 * input_scale * weight_scale).astype(np.int32)
                    node.input_tensors[2].data = bias_s32
                    node.input_tensors[2].dtype = 's32'

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Special file for triggering multiple instances for benchmarking."""
import argparse
import time
import torch
from collections import UserDict
from neural_compressor.utils import logger
from intel_extension_for_transformers.optimization import OptimizedModel


def torch_inference(model, data):
    start_time = time.time()
    with torch.no_grad():
        if isinstance(data, dict) or isinstance(data, UserDict):
            model(**data)
        elif isinstance(data, list) or isinstance(data, tuple):
            model(*data)
        else:
            model(data)
    return time.time() - start_time


def get_latency_per_iter(model, data, warmup, iters):
    time_usage = 0
    for i in range(iters):
        if i < warmup:
            torch_inference(model, data)
        else:
            time_usage += torch_inference(model, data)
    return time_usage / (iters - warmup)


if __name__ == "__main__":
    logger.info("Evaluating performance:")
    parser = argparse.ArgumentParser(description="Evaluate performance")
    parser.add_argument("--model", type=str, help="input model")
    parser.add_argument("--data", type=str, help="example inputs")
    parser.add_argument("--batch_size", default=1, type=int, help="the batch size of input data")
    parser.add_argument("--warmup", default=5, type=int, help="warmup iterations")
    parser.add_argument("--iters", default=20, type=int, help="total iterations")
    parser.add_argument("--torchscript", action='store_true', help="the model is torchscript")
    parser.add_argument("--generate", action='store_true', help="evaluate model.generate performance")
    parser.add_argument("--from_pretrain", action='store_true', help="load model ")
    parser.add_argument("--enable_ipex", action='store_true', help="load model ")
    args = parser.parse_args()
    if args.enable_ipex:
        import intel_extension_for_pytorch as ipex
    if args.torchscript:
        model = torch.jit.load(args.model)
    else:
        if args.from_pretrain:
            model = OptimizedModel.from_pretrained(args.model)
        else:
            model = torch.load(args.model)
    model.eval()
    if args.generate:
        model = model.generate
    data = torch.load(args.data)
    time_usage = get_latency_per_iter(model, data, args.warmup, args.iters)
    latency = time_usage / args.batch_size
    throughput = args.batch_size / time_usage
    logger.info("Batch size: {}".format(args.batch_size))
    logger.info("Latency: {:.4f} second/sample".format(latency))
    logger.info("Throughput: {:.4f} samples/second".format(throughput))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine onnx utils."""

import numpy as np
from .ops.tensor import Tensor
from . import graph_utils as util
import copy

node_names = {}

def get_node_name(node):
    if node in node_names.keys():
        return node_names[node]
    else:
        name = node.kind() + "_" + str(len(node_names.keys()))
        node_names[node] = name
        return name

op_maps = {'aten::softmax': 'Softmax', 'prim::Constant': 'Constant',
           'aten::linear': 'InnerProduct', 'aten::slice': 'Slice', 'aten::unsqueeze': 'Unsqueeze',
           'aten::embedding': 'Gather', 'aten::where': 'Where', 'aten::matmul': 'Matmul', 'aten::gelu': 'Gelu',
           'aten::layer_norm': 'LayerNorm', 'aten::size': 'Shape', 'aten::view': 'View',
           'aten::permute': 'Reorder', 'aten::transpose': 'Reorder', 'aten::repeat': 'Repeat',
           'aten::select': 'Slice', 'aten::tanh': 'Tanh', 'aten::arange': 'Arange', 'aten::pow': 'Pow',
           'aten::reciprocal': 'Reciprocal', 'aten::einsum': 'Einsum', 'aten::sin': 'Sin', 'aten::cos': 'Cos',
           'aten::neg': 'Neg', 'aten::stack': 'Stack', 'aten::flatten': 'Flatten', 'aten::cat': 'Concat',
           'aten::rsub': 'Rsub', 'aten::mul': 'Mul', 'aten::add': 'Add', 'aten::add_': 'Add', 'aten::div': 'Div',
           'aten::sub': 'Sub', 'aten::gt': 'Greater', 'aten::lt': 'Less', 'aten::eq': 'Equal', 'aten::ne': 'NotEqual',
           'aten::quantize_per_tensor': 'Quantize', 'aten::dequantize': 'Dequantize',
           'aten::padding_sequence': 'PaddingSequence', 'aten::expand': 'Expand', 'aten::masked_fill': 'Masked_fill',
           'aten::floor_divide': 'Floor_divide', 'aten::max': 'Max', 'aten::mean': 'Mean', 'aten::reshape': 'Reshape',
           'aten::rsqrt': 'Rsqrt', 'aten::silu': 'Silu',
           'prim::ListUnpack': 'ListUnpack', 'prim::ListConstruct': 'ListConstruct',
           'prim::TupleUnpack': 'TupleUnpack', 'prim::TupleConstruct': 'TupleConstruct'}

def torch_extract_operator(node, model, nodes_dict, engine_graph=None):
    """Decorate the operator in Torch.

    Args:
        node: NodeProto
        model: Torchscript
        nodes_dict: dict, return value from graph_node_names_details
        tf_dtypes: dict, for get the dtype string

    Returns:
        op_type: node op type
        input_tensors: Tensor list, contains the node input tensors info
        output_tensors: Tensor list, contains the node output tensor info
    """
    op_type = node.kind()
    input_tensors = []
    output_tensors = []

    for i in range(node.inputsSize()):
        in_val = node.inputsAt(i)
        input_tensor_name = in_val.debugName()
        pre_node = in_val.node()
        source_ops = []
        source_ops.append(get_node_name(pre_node))
        if input_tensor_name in nodes_dict:
            input_tensor = nodes_dict[input_tensor_name]
            # input_tensor = copy.deepcopy(nodes_dict[input_tensor_name])
            # input_tensor.source_ops=source_ops
            # input_tensor.dest_op=[get_node_name(node)]
        else:
            input_tensor = Tensor(name=input_tensor_name,
                source_op=source_ops,
                dest_op=[get_node_name(node)],
                shape=None,
                data=None,
                dtype=None
                )
        input_tensors.append(input_tensor)

    for i in range(node.outputsSize()):
        out_val = node.outputsAt(i)
        output_tensor_name = out_val.debugName()
        dest_ops = []
        for val_user in out_val.uses():
            next_node = val_user.user
            if next_node.kind() != 'prim::Return':
                dest_ops.append(get_node_name(next_node))
        output_tensor = Tensor(name=output_tensor_name,
                source_op=[get_node_name(node)],
                dest_op=dest_ops,
                shape=None,
                data=None,
                dtype=None
                )
        output_tensors.append(output_tensor)

    return op_maps[op_type], input_tensors, output_tensors




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine operator mapping file."""

from .op import Operator, operator_registry
from .tensor import Tensor


# get the matmul op scale, in bert_mlperf_int8.pb model, this op has several outputs
# but the other output tensors has empty shape except the first one, and don't be used by other ops
# so just keep the first one as the output_tensor
@operator_registry(operator_type='QuantizeV2')
class QuantizeV2(Operator):
    """Parse the QuantizeV2 operator to the neural engine."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def set_attr(self, framework, node):
        """Extract the node attr from tensorflow."""
        self._op_type = 'Quantize'
        if framework == 'tensorflow':
            self._attr['output_dtype'] = 'u8'



import copy
import mlflow
import numpy as np
import os
import shutil
import torch.utils.data as data
import unittest
from datasets import load_dataset, load_metric
from intel_extension_for_transformers.optimization import (
    DistillationConfig,
    DistillationCriterionMode,
    metrics,
    OptimizedModel,
    NoTrainerOptimizer
)
from intel_extension_for_transformers.optimization.trainer import NLPTrainer
from intel_extension_for_transformers.optimization.distillation import Criterion
from transformers import (
    AutoModelForSequenceClassification,
    AutoTokenizer,
    set_seed,
)

os.environ["WANDB_DISABLED"] = "true"


class TestDistillation(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        set_seed(42)
        self.model = AutoModelForSequenceClassification.from_pretrained(
            'distilbert-base-uncased'
        )
        self.teacher_model = AutoModelForSequenceClassification.from_pretrained(
            'distilbert-base-uncased-finetuned-sst-2-english'
        )
        self.optimizer = NoTrainerOptimizer(self.model)
        raw_datasets = load_dataset("glue", "sst2")["validation"]
        tokenizer = AutoTokenizer.from_pretrained("distilbert-base-uncased")
        def preprocess_function(examples):
            # Tokenize the texts
            args = (
                (examples['sentence'],)
            )
            result = tokenizer(*args, padding=True, max_length=64, truncation=True)
            return result
        raw_datasets = raw_datasets.map(
            preprocess_function, batched=True, load_from_cache_file=True
        )
        eval_dataset = raw_datasets.select(range(30))
        self.dataset = eval_dataset

    @classmethod
    def tearDownClass(self):
        shutil.rmtree('./tmp_trainer', ignore_errors=True)
        shutil.rmtree('./distilled_model', ignore_errors=True)

    def test_fx_model_distil(self):
        metric = load_metric("accuracy")
        def compute_metrics(p):
            preds = p.predictions
            preds = np.argmax(preds, axis=1)
            return metric.compute(predictions=preds, references=p.label_ids)
        origin_weight = copy.deepcopy(self.model.classifier.weight)
        for mode in DistillationCriterionMode:
            print("Distillation approach:", mode.value)
            self.trainer = NLPTrainer(
                model=copy.deepcopy(self.model),
                train_dataset=self.dataset,
                eval_dataset=self.dataset,
                compute_metrics=compute_metrics,
            )
            metric_ = metrics.Metric(name="eval_accuracy")
            criterion = Criterion(
                name='IntermediateLayersLoss',
                layer_mappings=[['classifier', 'classifier']],
                loss_types=['MSE'],
                loss_weight_ratio=[1.0],
                add_origin_loss=False
            ) if mode.value == "IntermediateLayersKnowledgeDistillationLoss" else None
            distillation_conf = DistillationConfig(metrics=metric_, criterion=criterion)
            distilled_model = self.trainer.distill(
                distillation_config=distillation_conf, teacher_model=self.teacher_model
            )
            # By default, model will be saved in tmp_trainer dir.
            self.trainer.save_model('./distilled_model')
            loaded_model = OptimizedModel.from_pretrained(
                './distilled_model',
            )
            distilled_weight = copy.deepcopy(distilled_model.classifier.weight)
            loaded_weight = copy.deepcopy(loaded_model.classifier.weight)
            # check distilled model
            self.assertTrue((distilled_weight != origin_weight).any())
            # check loaded model
            self.assertTrue((distilled_weight == loaded_weight).all())
            mlflow.end_run()

    def test_functional_distil(self):
        def eval_func(model):
            return 1

        def train_func(model):
            return model

        self.trainer = NLPTrainer(self.model)

        distillation_conf = DistillationConfig()
        self.trainer.distill(distillation_conf,
                           teacher_model=self.teacher_model,
                           provider="inc",
                           train_func = train_func,
                           eval_func = eval_func,)

    def test_no_trainer_distill(self):
        def eval_func(model):
            return 1

        def train_func(model):
            return model

        distillation_conf = DistillationConfig()
        self.optimizer.eval_func = eval_func
        self.optimizer.train_func = train_func
        self.optimizer.distill(distillation_conf,
                           teacher_model=self.teacher_model,
                           provider="inc",
                           train_func = train_func,
                           eval_func = eval_func,)
if __name__ == "__main__":
    unittest.main()
